import os
from pathlib import Path

from git import Repo

from flexlate.add_mode import AddMode, get_expanded_out_root
from flexlate.branch_update import modify_files_via_branches_and_temp_repo
from flexlate.config_manager import (
    ConfigManager,
    determine_config_path_from_roots_and_add_mode,
)
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.exc import CannotRemoveAppliedTemplateException
from flexlate.ext_git import (
    assert_repo_is_in_clean_state,
    update_local_branches_from_remote_without_checkout,
)
from flexlate.path_ops import (
    location_relative_to_new_parent,
)
from flexlate.render.multi import MultiRenderer
from flexlate.styles import console, styled, INFO_STYLE, print_styled, SUCCESS_STYLE
from flexlate.transactions.transaction import (
    FlexlateTransaction,
    create_transaction_commit_message,
)
from flexlate.update.main import Updater


class Remover:
    def remove_template_source(
        self,
        repo: Repo,
        template_name: str,
        transaction: FlexlateTransaction,
        out_root: Path = Path("."),
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        add_mode: AddMode = AddMode.LOCAL,
        remote: str = "origin",
        config_manager: ConfigManager = ConfigManager(),
    ):
        assert_repo_is_in_clean_state(repo)
        if repo.working_dir is None:
            raise ValueError("repo working dir should not be none")

        config_path = determine_config_path_from_roots_and_add_mode(
            out_root, Path(repo.working_dir), add_mode
        )
        project_root = Path(repo.working_dir)

        with console.status(styled("Removing template source...", INFO_STYLE)):
            print_styled(f"Removing template source {template_name}", INFO_STYLE)
            if add_mode == AddMode.USER:
                # No need to use git if was added for user
                config_manager.remove_template_source(
                    template_name, config_path=config_path, project_root=project_root
                )
                return

            commit_message = create_transaction_commit_message(
                _remove_template_source_commit_message(
                    template_name, out_root, Path(repo.working_dir)
                ),
                transaction,
            )

            # Ensure that all flexlate branches are up to date from remote before working on them
            update_local_branches_from_remote_without_checkout(
                repo,
                [
                    base_merged_branch_name,
                    merged_branch_name,
                    base_template_branch_name,
                    template_branch_name,
                ],
                remote=remote,
            )

            modify_files_via_branches_and_temp_repo(
                lambda temp_path: config_manager.remove_template_source(
                    template_name,
                    location_relative_to_new_parent(
                        config_path, project_root, temp_path, Path(os.getcwd())
                    ),
                    project_root=temp_path,
                ),
                repo,
                commit_message,
                out_root,
                merged_branch_name=merged_branch_name,
                base_merged_branch_name=base_merged_branch_name,
                template_branch_name=template_branch_name,
                base_template_branch_name=base_template_branch_name,
                remote=remote,
            )
            print_styled(
                f"Successfully removed template source {template_name}", SUCCESS_STYLE
            )

    def remove_applied_template_and_output(
        self,
        repo: Repo,
        template_name: str,
        transaction: FlexlateTransaction,
        out_root: Path = Path("."),
        add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        remote: str = "origin",
        config_manager: ConfigManager = ConfigManager(),
        updater: Updater = Updater(),
        renderer: MultiRenderer = MultiRenderer(),
    ):
        assert_repo_is_in_clean_state(repo)
        if repo.working_dir is None:
            raise ValueError("repo working dir should not be none")

        project_root = Path(repo.working_dir)
        template = config_manager.get_template_by_name(
            template_name, project_root=project_root
        )

        # Determine location of config
        # Need to get renderable to render path in case it is templated
        # TODO: can use get all renderables
        renderables = config_manager.get_renderables_for_updates(
            config_manager.get_no_op_updates(project_root=project_root),
            project_root=project_root,
        )
        if len(renderables) == 0:
            raise CannotRemoveAppliedTemplateException(
                f"Cannot find any applied template with name {template_name} "
                f"because there are no applied templates"
            )
        renderable = renderables[0]
        new_relative_out_root = Path(
            renderer.render_string(
                str(template.render_relative_root_in_output), renderable
            )
        )
        full_local_config_out_root = out_root / new_relative_out_root
        config_path = determine_config_path_from_roots_and_add_mode(
            full_local_config_out_root, project_root, add_mode
        )

        expanded_out_root = get_expanded_out_root(
            out_root, project_root, template.render_relative_root_in_output, add_mode
        )

        with console.status(styled("Removing applied template...", INFO_STYLE)):
            print_styled(
                f"Removing applied template {template_name} at {expanded_out_root.resolve()}",
                INFO_STYLE,
            )
            if add_mode == AddMode.USER:
                # No need to commit config changes for user
                config_manager.remove_applied_template(
                    template_name,
                    config_path,
                    project_root=project_root,
                    out_root=expanded_out_root,
                    orig_project_root=project_root,
                )
            else:
                # Commit changes for local and project
                commit_message = create_transaction_commit_message(
                    _remove_applied_template_commit_message(
                        template_name, out_root, Path(repo.working_dir)
                    ),
                    transaction,
                )

                # Ensure that all flexlate branches are up to date from remote before working on them
                update_local_branches_from_remote_without_checkout(
                    repo,
                    [
                        base_merged_branch_name,
                        merged_branch_name,
                        base_template_branch_name,
                        template_branch_name,
                    ],
                    remote=remote,
                )

                modify_files_via_branches_and_temp_repo(
                    lambda temp_path: config_manager.remove_applied_template(
                        template_name,
                        location_relative_to_new_parent(
                            config_path, project_root, temp_path, Path(os.getcwd())
                        ),
                        project_root=temp_path,
                        out_root=expanded_out_root,
                        orig_project_root=project_root,
                    ),
                    repo,
                    commit_message,
                    out_root,
                    merged_branch_name=merged_branch_name,
                    base_merged_branch_name=base_merged_branch_name,
                    template_branch_name=template_branch_name,
                    base_template_branch_name=base_template_branch_name,
                    remote=remote,
                )

            updater.update(
                repo,
                [],
                transaction,
                merged_branch_name=merged_branch_name,
                base_merged_branch_name=base_merged_branch_name,
                template_branch_name=template_branch_name,
                base_template_branch_name=base_template_branch_name,
                no_input=True,
                remote=remote,
                renderer=renderer,
                config_manager=config_manager,
            )
            print_styled(
                f"Successfully removed applied template {template_name} at {expanded_out_root.resolve()}",
                SUCCESS_STYLE,
            )


def _remove_template_source_commit_message(
    template_name: str, out_root: Path, project_root: Path
) -> str:
    relative_path = out_root.absolute().relative_to(project_root)
    return f"Removed template source {template_name} from {relative_path}"


def _remove_applied_template_commit_message(
    template_name: str, out_root: Path, project_root: Path
) -> str:
    relative_path = out_root.absolute().relative_to(project_root)
    return f"Removed template {template_name} from {relative_path}"



from git import Repo

from flexlate.path_ops import change_directory_to
from tests.config import GENERATED_REPO_DIR
from tests.integration.fixtures.repo import *
from tests.integration.cli.fxt import fxt


def test_update_returns_code_0_for_update_without_conflict(
    repo_with_copier_remote_version_one_and_no_target_version: Repo,
):
    with change_directory_to(GENERATED_REPO_DIR):
        result = fxt(["update", "--no-input", "--abort"])
    assert result.exit_code == 0


def test_update_returns_code_1_for_template_conflict_with_abort(
    repo_with_copier_remote_version_one_no_target_version_and_will_have_a_conflict_on_update: Repo,
):
    with change_directory_to(GENERATED_REPO_DIR):
        result = fxt(["update", "--no-input", "--abort"])
    assert result.exit_code == 1


def test_update_returns_code_1_for_template_conflict_with_abort_and_no_cleanup(
    repo_with_copier_remote_version_one_no_target_version_and_will_have_a_conflict_on_update: Repo,
):
    with change_directory_to(GENERATED_REPO_DIR):
        result = fxt(["update", "--no-input", "--abort", "--no-cleanup"])
    assert result.exit_code == 1



import os
from pathlib import Path
from typing import Callable, Union, List, Optional

from git import Repo, repo, Head

from flexlate.cli_utils import confirm_user
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.ext_git import (
    temp_repo_that_pushes_to_branch,
    stage_and_commit_all,
    fast_forward_branch_without_checkout,
    checked_out_template_branch,
    merge_branch_into_current,
    abort_merge,
    reset_branch_to_commit_without_checkout,
    repo_has_merge_conflicts,
    get_branch_sha,
)
from flexlate.path_ops import make_func_that_creates_cwd_and_out_root_before_running
from flexlate.styles import (
    print_styled,
    ACTION_REQUIRED_STYLE,
    styled,
    QUESTION_STYLE,
    ALERT_STYLE,
)
from flexlate.transactions.transaction import (
    FlexlateTransaction,
    reset_last_transaction,
)


def modify_files_via_branches_and_temp_repo(
    file_operation: Callable[[Path], None],
    repo: Repo,
    commit_message: str,
    out_root: Path,
    merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
    base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
    template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    remote: str = "origin",
):
    cwd = os.getcwd()
    current_branch = repo.active_branch

    # Save the status of the flexlate branches. We may need to roll back to this state
    # if the user aborts the operation
    merged_branch_sha = get_branch_sha(repo, merged_branch_name)
    template_branch_sha = get_branch_sha(repo, template_branch_name)

    make_dirs_add_operation = make_func_that_creates_cwd_and_out_root_before_running(
        out_root, file_operation
    )

    # Update the template only branch with the new template
    with temp_repo_that_pushes_to_branch(  # type: ignore
        repo,
        branch_name=template_branch_name,
        base_branch_name=base_template_branch_name,
        remote=remote,
    ) as temp_repo:
        make_dirs_add_operation(Path(temp_repo.working_dir))  # type: ignore
        stage_and_commit_all(temp_repo, commit_message)

    # Bring the change into the merged branch
    # Update with changes from the main repo
    fast_forward_branch_without_checkout(repo, merged_branch_name, current_branch.name)
    with checked_out_template_branch(
        repo, branch_name=merged_branch_name, base_branch_name=base_merged_branch_name
    ):
        # Update with the new template
        merge_branch_into_current(repo, template_branch_name)
        if repo_has_merge_conflicts(repo):
            aborted = prompt_to_fix_conflicts_and_reset_on_abort_return_aborted(
                repo,
                current_branch,
                merged_branch_sha,
                template_branch_sha,
                merged_branch_name,
                template_branch_name,
            )
            if aborted:
                return

    # Merge back into current branch
    merge_branch_into_current(repo, merged_branch_name)

    # Current working directory or out root may have been deleted if it was a remove operation
    # and there was nothing else in the folder (git does not save folders without files)
    ensure_exists_folders: List[Union[str, Path]] = [cwd, out_root]
    for folder in ensure_exists_folders:
        if not os.path.exists(folder):
            os.makedirs(folder)

    # Folder may have been deleted again while switching branches, so
    # need to set cwd again
    os.chdir(cwd)


def undo_transaction_in_flexlate_branches(
    repo: Repo,
    transaction: FlexlateTransaction,
    merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
    base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
    template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
    base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
):
    # Reset the template only branch to the appropriate commit
    with temp_repo_that_pushes_to_branch(  # type: ignore
        repo,
        branch_name=template_branch_name,
        base_branch_name=base_template_branch_name,
        force_push=True,
    ) as temp_repo:
        reset_last_transaction(
            temp_repo, transaction, merged_branch_name, template_branch_name
        )

    # Reset the merged template branch to the appropriate commit
    with temp_repo_that_pushes_to_branch(  # type: ignore
        repo,
        branch_name=merged_branch_name,
        base_branch_name=base_merged_branch_name,
        force_push=True,
    ) as temp_repo:
        reset_last_transaction(
            temp_repo, transaction, merged_branch_name, template_branch_name
        )


def abort_merge_and_reset_flexlate_branches(
    repo: Repo,
    current_branch: Head,
    merged_branch_sha: Optional[str] = None,
    template_branch_sha: Optional[str] = None,
    merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
    template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
):
    abort_merge(repo)
    current_branch.checkout()
    reset_branch_to_commit_without_checkout(repo, merged_branch_name, merged_branch_sha)
    reset_branch_to_commit_without_checkout(
        repo, template_branch_name, template_branch_sha
    )


def prompt_to_fix_conflicts_and_reset_on_abort_return_aborted(
    repo: Repo,
    current_branch: Head,
    merged_branch_sha: Optional[str],
    template_branch_sha: Optional[str],
    merged_branch_name: str,
    template_branch_name: str,
) -> bool:
    print_styled(
        "Repo has merge conflicts after update, please resolve them",
        ACTION_REQUIRED_STYLE,
    )
    user_fixed = confirm_user(styled("Conflicts fixed? n to abort", QUESTION_STYLE))
    if not user_fixed:
        print_styled(
            "Aborting update.",
            ALERT_STYLE,
        )
        abort_merge_and_reset_flexlate_branches(
            repo,
            current_branch,
            merged_branch_sha=merged_branch_sha,
            template_branch_sha=template_branch_sha,
            merged_branch_name=merged_branch_name,
            template_branch_name=template_branch_name,
        )
        return True
    return False


def get_flexlate_branch_name(repo: Repo, base_branch_name: str) -> str:
    current_branch = repo.active_branch.name
    return get_flexlate_branch_name_for_feature_branch(current_branch, base_branch_name)


def get_flexlate_branch_name_for_feature_branch(
    feature_branch: str, base_branch_name: str
) -> str:
    return f"{base_branch_name}-{feature_branch}"



import tempfile
from pathlib import Path
from typing import Tuple

from cookiecutter.config import get_user_config
from cookiecutter.generate import generate_context, generate_files
from cookiecutter.prompt import prompt_for_config

from flexlate.render.renderable import Renderable
from flexlate.render.specific.base import SpecificTemplateRenderer
from flexlate.template.cookiecutter import CookiecutterTemplate
from flexlate.template.types import TemplateType
from flexlate.template_data import TemplateData


class CookiecutterRenderer(SpecificTemplateRenderer[CookiecutterTemplate]):
    _template_cls = CookiecutterTemplate
    _template_type = TemplateType.COOKIECUTTER

    def render(
        self,
        renderable: Renderable[CookiecutterTemplate],
        no_input: bool = False,
    ) -> TemplateData:
        template = renderable.template
        config_dict = get_user_config()
        context_file = template.path / "cookiecutter.json"

        context = generate_context(
            context_file=context_file,
            default_context=config_dict["default_context"],
            extra_context=renderable.data,
        )
        context["cookiecutter"] = prompt_for_config(context, no_input)
        context["cookiecutter"]["_template"] = template.path

        generate_files(
            repo_dir=str(template.path),
            context=context,
            overwrite_if_exists=False,
            skip_if_file_exists=False,
            output_dir=str(renderable.out_root),
        )

        used_data = dict(context["cookiecutter"])
        used_data.pop("_template")

        return used_data

    def render_string(
        self,
        string: str,
        renderable: Renderable[CookiecutterTemplate],
    ) -> str:
        template = renderable.template
        config_dict = get_user_config()
        context_file = template.path / "cookiecutter.json"

        context = generate_context(
            context_file=context_file,
            default_context=config_dict["default_context"],
            extra_context=renderable.data,
        )
        context["cookiecutter"] = prompt_for_config(context, no_input=True)

        with tempfile.TemporaryDirectory() as temp_template_dir:
            context["cookiecutter"]["_template"] = temp_template_dir
            (
                cookiecutter_pre_folder_name,
                cookiecutter_post_folder_name,
            ) = _cookiecutter_before_and_after_render_string_temp_folder(context)
            temp_template_file_path = (
                Path(temp_template_dir) / cookiecutter_pre_folder_name / "temp.txt"
            )
            temp_template_file_path.parent.mkdir()
            temp_template_file_path.write_text(string)

            with tempfile.TemporaryDirectory() as temp_output_dir:
                generate_files(
                    repo_dir=temp_template_dir,
                    context=context,
                    overwrite_if_exists=False,
                    skip_if_file_exists=False,
                    output_dir=temp_output_dir,
                )
                temp_output_file_path = (
                    Path(temp_output_dir) / cookiecutter_post_folder_name / "temp.txt"
                )
                output = temp_output_file_path.read_text()

        return output


def _get_first_question_and_answer_from_context(context: dict) -> Tuple[str, str]:
    cookiecutter_context = context["cookiecutter"]
    for question, answer in cookiecutter_context.items():
        return question, answer
    raise ValueError("no questions in cookiecutter context, this is not expected")


def _cookiecutter_before_and_after_render_string_temp_folder(
    cookiecutter_context: dict,
) -> Tuple[str, str]:
    question, answer = _get_first_question_and_answer_from_context(cookiecutter_context)
    return "{{ cookiecutter." + question + " }}", answer



import json

import pytest
from git import Repo

from flexlate.config import FlexlateConfig
from flexlate.ext_git import stage_and_commit_all
from flexlate.main import Flexlate
from flexlate.path_ops import change_directory_to
from flexlate.user_config_manager import UserConfigManager
from tests.config import (
    GENERATED_REPO_DIR,
    COOKIECUTTER_REMOTE_URL,
    COOKIECUTTER_REMOTE_VERSION_1,
    COPIER_REMOTE_NAME,
    COPIER_REMOTE_URL,
    COPIER_REMOTE_VERSION_1,
)
from tests.fixtures.git import *

fxt = Flexlate()


@pytest.fixture
def repo_with_default_flexlate_project(repo_with_placeholder_committed: Repo) -> Repo:
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
    yield repo


@pytest.fixture
def repo_with_copier_remote_version_one(
    repo_with_default_flexlate_project: Repo,
) -> Repo:
    repo = repo_with_default_flexlate_project
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.add_template_source(
            COPIER_REMOTE_URL, target_version=COPIER_REMOTE_VERSION_1
        )
        # TODO: Allow updating template sources even when there are no applied templates
        #  It should not be necessary to add this template for the update cli integration tests to pass
        #  Need to rework the process of getting updates, right now it relies on there being applied templates
        fxt.apply_template_and_add(COPIER_REMOTE_NAME, no_input=True)
    yield repo


@pytest.fixture
def repo_with_copier_remote_version_one_and_no_target_version(
    repo_with_copier_remote_version_one: Repo,
) -> Repo:
    repo = repo_with_copier_remote_version_one
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.update_template_source_target_version(
            COPIER_REMOTE_NAME, None, project_path=GENERATED_REPO_DIR
        )
    yield repo


@pytest.fixture
def repo_with_copier_remote_version_one_no_target_version_and_will_have_a_conflict_on_update(
    repo_with_copier_remote_version_one_and_no_target_version: Repo,
) -> Repo:
    repo = repo_with_copier_remote_version_one_and_no_target_version
    # Reformat the flexlate config to cause a conflict
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config_data = json.loads(config_path.read_text())
    config_path.write_text(json.dumps(config_data, indent=4))
    stage_and_commit_all(repo, "Reformat flexlate config to cause a conflict on update")
    yield repo



# Adapted from https://stackoverflow.com/a/54477583/6276321
import hashlib
from _hashlib import HASH as Hash
from pathlib import Path
from typing import Union


def md5_update_from_file(filename: Union[str, Path], hash: Hash) -> Hash:
    assert Path(filename).is_file()
    with open(str(filename), "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
    return hash


def md5_file(filename: Union[str, Path]) -> str:
    return str(md5_update_from_file(filename, hashlib.md5()).hexdigest())  # type: ignore


def md5_update_from_dir(directory: Union[str, Path], hash: Hash) -> Hash:
    assert Path(directory).is_dir()
    for path in sorted(Path(directory).iterdir(), key=lambda p: str(p).lower()):
        hash.update(path.name.encode())
        if path.is_file():
            hash = md5_update_from_file(path, hash)
        elif path.is_dir():
            hash = md5_update_from_dir(path, hash)
    return hash


def md5_dir(directory: Union[str, Path]) -> str:
    return str(md5_update_from_dir(directory, hashlib.md5()).hexdigest())  # type: ignore



import os
import shutil
import tempfile
import time
from copy import deepcopy
from pathlib import Path
from typing import Optional

from git import Repo
from rich.prompt import Prompt

from flexlate.add_mode import AddMode, get_expanded_out_root
from flexlate.branch_update import modify_files_via_branches_and_temp_repo
from flexlate.config import TemplateSource
from flexlate.config_manager import (
    ConfigManager,
    determine_config_path_from_roots_and_add_mode,
)
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.exc import TemplateSourceWithNameAlreadyExistsException
from flexlate.ext_git import (
    assert_repo_is_in_clean_state,
    stage_and_commit_all,
    update_local_branches_from_remote_without_checkout,
)
from flexlate.path_ops import (
    location_relative_to_new_parent,
)
from flexlate.render.multi import MultiRenderer
from flexlate.styles import SUCCESS_STYLE, INFO_STYLE, console, styled, print_styled
from flexlate.syncer import Syncer
from flexlate.template.base import Template
from flexlate.template_data import TemplateData
from flexlate.transactions.transaction import (
    FlexlateTransaction,
    create_transaction_commit_message,
)
from flexlate.update.main import Updater
from flexlate.update.template import TemplateUpdate


class Adder:
    def add_template_source(
        self,
        repo: Repo,
        template: Template,
        transaction: FlexlateTransaction,
        target_version: Optional[str] = None,
        out_root: Path = Path("."),
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        add_mode: AddMode = AddMode.LOCAL,
        remote: str = "origin",
        config_manager: ConfigManager = ConfigManager(),
    ):
        assert_repo_is_in_clean_state(repo)
        if repo.working_dir is None:
            raise ValueError("repo working dir should not be none")

        config_path = determine_config_path_from_roots_and_add_mode(
            out_root, Path(repo.working_dir), add_mode
        )
        project_root = Path(repo.working_dir)

        if config_manager.template_source_exists(
            template.name, project_root=project_root
        ):
            raise TemplateSourceWithNameAlreadyExistsException(
                f"There is an existing template source with the name {template.name}. "
                f"To add this one, give it a custom name or remove the existing one"
            )

        if not template.path.is_absolute() and not (
            config_path.parent.resolve() == Path(os.getcwd())
        ):
            # Relative path, if the config output location is different than the
            # current directory, need to adjust the path to be relative to
            # the config output location

            # Don't overwrite existing template
            template = deepcopy(template)
            # Make template path relative to its config file rather than current directory
            template.path = Path(
                os.path.relpath(template.path.resolve(), config_path.parent.resolve())
            )

        with console.status(print_styled("Adding template source...", INFO_STYLE)):
            print_styled(
                f"Adding template source {template.name} from {template.git_url or template.path}",
                INFO_STYLE,
            )

            if add_mode == AddMode.USER:
                # No need to use git if adding for user
                config_manager.add_template_source(
                    template,
                    config_path,
                    target_version=target_version,
                    project_root=Path(repo.working_dir),  # type: ignore
                )
                return

            commit_message = create_transaction_commit_message(
                _add_template_source_commit_message(
                    template, out_root, Path(repo.working_dir)
                ),
                transaction,
            )

            # Ensure that all flexlate branches are up to date from remote before working on them
            update_local_branches_from_remote_without_checkout(
                repo,
                [
                    base_merged_branch_name,
                    merged_branch_name,
                    base_template_branch_name,
                    template_branch_name,
                ],
                remote=remote,
            )

            # Local or project config, add in git
            modify_files_via_branches_and_temp_repo(
                lambda temp_path: config_manager.add_template_source(
                    template,
                    location_relative_to_new_parent(
                        config_path, project_root, temp_path, Path(os.getcwd())
                    ),
                    target_version=target_version,
                    project_root=temp_path,
                ),
                repo,
                commit_message,
                out_root,
                merged_branch_name=merged_branch_name,
                base_merged_branch_name=base_merged_branch_name,
                template_branch_name=template_branch_name,
                base_template_branch_name=base_template_branch_name,
                remote=remote,
            )

            print_styled(
                f"Sucessfully added template source {template.name}", INFO_STYLE
            )

    def apply_template_and_add(
        self,
        repo: Repo,
        template: Template,
        transaction: FlexlateTransaction,
        data: Optional[TemplateData] = None,
        out_root: Path = Path("."),
        add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        no_input: bool = False,
        remote: str = "origin",
        config_manager: ConfigManager = ConfigManager(),
        updater: Updater = Updater(),
        renderer: MultiRenderer = MultiRenderer(),
    ):
        assert_repo_is_in_clean_state(repo)
        if repo.working_dir is None:
            raise ValueError("repo working dir should not be none")

        project_root = Path(repo.working_dir)
        full_local_config_out_root = out_root / template.render_relative_root_in_output
        config_path = determine_config_path_from_roots_and_add_mode(
            full_local_config_out_root, project_root, add_mode
        )
        expanded_out_root = get_expanded_out_root(
            out_root, project_root, template.render_relative_root_in_output, add_mode
        )

        # Get path it is being rendered at to display to user
        render_path = expanded_out_root
        if not expanded_out_root.is_absolute():
            render_path = (project_root / expanded_out_root).resolve()

        print_styled(
            f"Applying template {template.name} to {render_path}",
            INFO_STYLE,
        )

        if add_mode == AddMode.USER:
            # No need to commit config changes for user
            config_manager.add_applied_template(
                template,
                config_path,
                add_mode,
                data=data,
                project_root=project_root,
                out_root=expanded_out_root,
            )
        else:
            # Commit changes for local and project
            commit_message = create_transaction_commit_message(
                _add_template_commit_message(
                    template, out_root, Path(repo.working_dir)
                ),
                transaction,
            )

            # Ensure that all flexlate branches are up to date from remote before working on them
            update_local_branches_from_remote_without_checkout(
                repo,
                [
                    base_merged_branch_name,
                    merged_branch_name,
                    base_template_branch_name,
                    template_branch_name,
                ],
                remote=remote,
            )

            modify_files_via_branches_and_temp_repo(
                lambda temp_path: config_manager.add_applied_template(
                    template,
                    location_relative_to_new_parent(
                        config_path, project_root, temp_path, Path(os.getcwd())
                    ),
                    add_mode,
                    data=data,
                    project_root=temp_path,
                    out_root=expanded_out_root,
                ),
                repo,
                commit_message,
                out_root,
                merged_branch_name=merged_branch_name,
                base_merged_branch_name=base_merged_branch_name,
                template_branch_name=template_branch_name,
                base_template_branch_name=base_template_branch_name,
                remote=remote,
            )
        template_update = TemplateUpdate(
            template=template,
            config_location=config_path,
            index=config_manager.get_num_applied_templates_in_child_config(
                config_path, project_root=project_root
            )
            - 1,
            data=data,
        )
        updater.update(
            repo,
            [template_update],
            transaction,
            merged_branch_name=merged_branch_name,
            base_merged_branch_name=base_merged_branch_name,
            template_branch_name=template_branch_name,
            base_template_branch_name=base_template_branch_name,
            no_input=no_input,
            full_rerender=False,
            remote=remote,
            renderer=renderer,
            config_manager=config_manager,
        )
        print_styled(
            f"Successfully applied template {template.name} to {render_path}",
            SUCCESS_STYLE,
        )

    def init_project_and_add_to_branches(
        self,
        repo: Repo,
        default_add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        user: bool = False,
        remote: str = "origin",
        config_manager: ConfigManager = ConfigManager(),
    ):
        assert_repo_is_in_clean_state(repo)
        if repo.working_dir is None:
            raise ValueError("repo working dir should not be none")

        path = Path(repo.working_dir)

        with console.status(styled("Initializing...", INFO_STYLE)):
            print_styled(
                f"Initializing flexlate project with default add mode "
                f"{default_add_mode.value} and user={user} in "
                f"{Path(repo.working_dir).resolve()}",
                INFO_STYLE,
            )

            if user:
                # Simply init the project for the user
                config_manager.add_project(
                    path=path,
                    default_add_mode=default_add_mode,
                    user=user,
                    merged_branch_name=merged_branch_name,
                    template_branch_name=template_branch_name,
                    remote=remote,
                )
                return

            # Config resides in project, so add it via branches
            modify_files_via_branches_and_temp_repo(
                lambda temp_path: config_manager.add_project(
                    path=temp_path,
                    default_add_mode=default_add_mode,
                    user=user,
                    merged_branch_name=merged_branch_name,
                    template_branch_name=template_branch_name,
                    remote=remote,
                ),
                repo,
                "Initialized flexlate project",
                path,
                merged_branch_name=merged_branch_name,
                base_merged_branch_name=merged_branch_name,
                template_branch_name=template_branch_name,
                base_template_branch_name=template_branch_name,
                remote=remote,
            )

        print_styled(
            f"Finished initializing flexlate project",
            SUCCESS_STYLE,
        )

    def init_project_from_template_source_path(
        self,
        template: Template,
        transaction: FlexlateTransaction,
        path: Path = Path("."),
        target_version: Optional[str] = None,
        data: Optional[TemplateData] = None,
        default_folder_name: str = "project",
        default_add_mode: AddMode = AddMode.LOCAL,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        remote: str = "origin",
        no_input: bool = False,
        config_manager: ConfigManager = ConfigManager(),
        updater: Updater = Updater(),
        renderer: MultiRenderer = MultiRenderer(),
        syncer: Syncer = Syncer(),
    ) -> str:
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = Path(temp_dir)

            repo = Repo.init(temp_dir)

            def sync():
                syncer.sync_local_changes_to_flexlate_branches(
                    repo,
                    transaction,
                    merged_branch_name=merged_branch_name,
                    base_merged_branch_name=merged_branch_name,
                    template_branch_name=template_branch_name,
                    base_template_branch_name=template_branch_name,
                    no_input=True,
                    remote=remote,
                    updater=updater,
                    renderer=renderer,
                    config_manager=config_manager,
                )

            temp_file = temp_path / "README.md"
            temp_file.touch()
            stage_and_commit_all(repo, "Initial commit")
            self.init_project_and_add_to_branches(
                repo,
                default_add_mode=default_add_mode,
                merged_branch_name=merged_branch_name,
                template_branch_name=template_branch_name,
                user=False,
                remote=remote,
                config_manager=config_manager,
            )
            self.add_template_source(
                repo,
                template,
                transaction,
                target_version=target_version,
                out_root=temp_path,
                merged_branch_name=merged_branch_name,
                template_branch_name=template_branch_name,
                add_mode=AddMode.LOCAL,
                remote=remote,
                config_manager=config_manager,
            )
            template = config_manager.get_template_by_name(
                template.name, project_root=temp_path
            )
            self.apply_template_and_add(
                repo,
                template,
                transaction,
                data=data,
                out_root=temp_path,
                add_mode=AddMode.LOCAL,
                merged_branch_name=merged_branch_name,
                template_branch_name=template_branch_name,
                no_input=no_input,
                remote=remote,
                config_manager=config_manager,
                updater=updater,
                renderer=renderer,
            )
            if template.render_relative_root_in_output == Path("."):
                output_folder = temp_path
                folder_name = default_folder_name
                if not no_input:
                    folder_name = Prompt.ask(
                        "Output folder name? (this template does not provide one)",
                        default=default_folder_name,
                    )
                if not len(temp_file.read_text()):
                    os.remove(temp_file)
                    stage_and_commit_all(repo, "Remove temporary file")
                final_temp_config_path = temp_path / "flexlate.json"
            else:
                # Output renders in a subdirectory. Find that directory
                renderables = config_manager.get_all_renderables(project_root=temp_path)
                if len(renderables) != 1:
                    raise ValueError(
                        "there should only be one renderable in the init-from process"
                    )
                renderable = renderables[0]
                new_relative_out_root = Path(
                    renderer.render_string(
                        str(template.render_relative_root_in_output), renderable
                    )
                )
                output_folder = (temp_path / new_relative_out_root).resolve()
                folder_name = new_relative_out_root.name

                # Move template source and project config into output directory
                orig_config_path = temp_path / "flexlate.json"
                final_temp_config_path = (
                    temp_path / new_relative_out_root / "flexlate.json"
                )
                config_manager.move_template_source(
                    template.name,
                    orig_config_path,
                    final_temp_config_path,
                    project_root=temp_path,
                )

                orig_project_config_path = temp_path / "flexlate-project.json"
                new_project_config_path = (
                    temp_path / new_relative_out_root / "flexlate-project.json"
                )
                shutil.move(str(orig_project_config_path), str(new_project_config_path))

                # Move .git folder into output directory
                git_folder = temp_path / ".git"
                new_git_folder = temp_path / new_relative_out_root / ".git"
                shutil.move(str(git_folder), str(new_git_folder))
                # Reassign repo now that it has moved
                repo = Repo(temp_path / new_relative_out_root)

                stage_and_commit_all(
                    repo, "Move flexlate config and remove temporary file"
                )
                sync()

            final_out_path = path / folder_name
            shutil.copytree(output_folder, final_out_path)
            repo = Repo(final_out_path)

            # Now update template source path that was previously relative to temp directory
            if (
                template.git_url is None
                and not Path(template.template_source_path).is_absolute()
            ):

                def move_source_path_to_be_relative_to_destination(
                    source: TemplateSource,
                ):
                    source.path = str(
                        os.path.relpath(
                            (
                                final_temp_config_path.parent / Path(source.path)
                            ).resolve(),
                            final_out_path,
                        )
                    )

                config_manager.update_template_sources(
                    [template.name],
                    move_source_path_to_be_relative_to_destination,
                    project_root=final_out_path,
                )
                stage_and_commit_all(
                    repo,
                    "Move template source path to match permanent destination of project",
                )
                sync()

            return folder_name


def _add_template_commit_message(
    template: Template, out_root: Path, project_root: Path
) -> str:
    relative_path = out_root.absolute().relative_to(project_root)
    return f"Applied template {template.name} to {relative_path}"


def _move_applied_template_config_message(
    template: Template, out_root: Path, project_root: Path
) -> str:
    relative_path = out_root.absolute().relative_to(project_root)
    return f"Moved config for {template.name} to {relative_path}"


def _add_template_source_commit_message(
    template: Template, out_root: Path, project_root: Path
) -> str:
    relative_path = out_root.absolute().relative_to(project_root)
    return f"Added template source {template.name} to {relative_path}"



import os
from copy import deepcopy
from pathlib import Path
from typing import Sequence, Optional, List, Dict, Any

from git import Repo, GitCommandError
from rich.prompt import Confirm

from flexlate.branch_update import (
    abort_merge_and_reset_flexlate_branches,
    prompt_to_fix_conflicts_and_reset_on_abort_return_aborted,
)
from flexlate.cli_utils import confirm_user
from flexlate.config_manager import ConfigManager
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.exc import (
    TriedToCommitButNoChangesException,
    MergeConflictsAndAbortException,
)
from flexlate.finder.multi import MultiFinder
from flexlate.logger import log
from flexlate.path_ops import (
    make_all_dirs,
    location_relative_to_new_parent,
    make_absolute_path_from_possibly_relative_to_another_path,
)
from flexlate.render.multi import MultiRenderer
from flexlate.render.renderable import Renderable
from flexlate.styles import (
    print_styled,
    ACTION_REQUIRED_STYLE,
    styled,
    QUESTION_STYLE,
    ALERT_STYLE,
    INFO_STYLE,
    SUCCESS_STYLE,
)
from flexlate.template.base import Template
from flexlate.ext_git import (
    stage_and_commit_all,
    merge_branch_into_current,
    checkout_template_branch,
    repo_has_merge_conflicts,
    assert_repo_is_in_clean_state,
    temp_repo_that_pushes_to_branch,
    fast_forward_branch_without_checkout,
    abort_merge,
    reset_branch_to_commit_without_checkout,
    get_branch_sha,
    restore_initial_commit_files,
    get_merge_conflict_diffs,
    update_local_branches_from_remote_without_checkout,
)
from flexlate.template_data import TemplateData, merge_data
from flexlate.transactions.transaction import (
    FlexlateTransaction,
    create_transaction_commit_message,
)
from flexlate.update.template import (
    TemplateUpdate,
    updates_with_updated_data,
)


class Updater:
    def update(
        self,
        repo: Repo,
        updates: Sequence[TemplateUpdate],
        transaction: FlexlateTransaction,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        no_input: bool = False,
        abort_on_conflict: bool = False,
        cleanup: bool = True,
        full_rerender: bool = True,
        remote: str = "origin",
        renderer: MultiRenderer = MultiRenderer(),
        config_manager: ConfigManager = ConfigManager(),
    ):
        assert_repo_is_in_clean_state(repo)
        if repo.working_dir is None:
            raise ValueError("repo working dir should not be none")

        project_root = Path(repo.working_dir)
        current_branch = repo.active_branch

        # Save the status of the flexlate branches. We may need to roll back to this state
        # if the user aborts the merge
        log.debug(
            f"Saving branch state before update for {merged_branch_name} and {template_branch_name}"
        )
        merged_branch_sha = get_branch_sha(repo, merged_branch_name)
        template_branch_sha = get_branch_sha(repo, template_branch_name)

        # Ensure that all flexlate branches are up to date from remote before working on them
        update_local_branches_from_remote_without_checkout(
            repo,
            [
                base_merged_branch_name,
                merged_branch_name,
                base_template_branch_name,
                template_branch_name,
            ],
            remote=remote,
        )

        # Prepare the template branch, this is the branch that stores only the template files
        # Create it from the initial commit if it does not exist
        cwd = Path(os.getcwd())
        with temp_repo_that_pushes_to_branch(  # type: ignore
            repo,
            branch_name=template_branch_name,
            base_branch_name=base_template_branch_name,
            delete_tracked_files=full_rerender,
            remote=remote,
        ) as temp_repo:
            temp_project_root = Path(temp_repo.working_dir)  # type: ignore
            log.debug(
                f"Working in a temporary repo at {temp_project_root} to update template branch {template_branch_name}"
            )
            temp_updates = _move_update_config_locations_to_new_parent(
                updates, project_root, temp_project_root
            )
            # On first update, don't use template source path. This means that
            # the template paths will be absolute, so they can be loaded even though we are
            # working in a temp directory
            config_manager.update_templates(
                temp_updates,
                project_root=temp_project_root,
                use_template_source_path=False,
            )
            orig_renderables = (
                config_manager.get_all_renderables(
                    relative_to=project_root, project_root=temp_project_root
                )
                if full_rerender
                else config_manager.get_renderables_for_updates(
                    updates, project_root=project_root
                )
            )
            if full_rerender:
                print_styled(
                    f"Syncing changes in flexlate configs to output", INFO_STYLE
                )
            else:
                print_styled(
                    f"Updating {len(orig_renderables)} applied templates", INFO_STYLE
                )

            prompt_set_renderables = (
                _copy_renderables_skipping_prompts_if_not_in_updates(
                    orig_renderables, updates, project_root=project_root
                )
            )
            renderables = _move_renderable_out_roots_to_new_parent(
                prompt_set_renderables,
                project_root,
                temp_project_root,
            )

            _create_cwd_and_directories_if_needed(
                cwd, [renderable.out_root for renderable in renderables]
            )
            updated_data = renderer.render(
                renderables, project_root=temp_project_root, no_input=no_input
            )
            new_updates = updates_with_updated_data(
                updates,
                updated_data,
                renderables,
                project_root=project_root,
                render_root=temp_project_root,
            )
            new_temp_updates = _move_update_config_locations_to_new_parent(
                new_updates, project_root, temp_project_root
            )
            # On second update, use template source path. This means that it will set the template
            # paths back to how they were originally (relative if needed), so that there will not be
            # unexpected changes from relative to absolute paths in the user configs
            config_manager.update_templates(
                new_temp_updates, project_root=temp_project_root
            )

            # For applied templates with local add mode, with corresponding template sources that
            # have a render_relative_root_in_output, may need to move the config
            config_manager.move_local_applied_templates_if_necessary(
                project_root=temp_project_root,
                orig_project_root=project_root,
                renderer=renderer,
            )

            # Add back initial commit files if they have not been rendered from a template
            if full_rerender:
                restore_initial_commit_files(temp_repo)

            commit_message = create_transaction_commit_message(
                _commit_message(renderables), transaction
            )
            try:
                stage_and_commit_all(temp_repo, commit_message)
            except GitCommandError as e:
                if "nothing to commit, working tree clean" in str(e):
                    # This is expected if user tries to run update or sync when not needed
                    raise TriedToCommitButNoChangesException(
                        "update did not make any new changes"
                    ) from e
                # Unexpected exception, raise it
                raise e
            log.debug("Leaving temp directory")

        log.debug(f"Updating merged branch {merged_branch_name}")
        # Now prepare the merged (output) branch, by merging the current
        # branch into it and then the template branch into it.
        # Update with changes from the main repo
        log.debug(
            f"Fast forwarding {merged_branch_name} based on {current_branch.name}"
        )
        fast_forward_branch_without_checkout(
            repo, merged_branch_name, current_branch.name
        )
        # Update with template changes
        log.debug(f"Merging {template_branch_name} into {merged_branch_name}")
        checkout_template_branch(repo, merged_branch_name, base_merged_branch_name)
        merge_branch_into_current(repo, template_branch_name)

        if repo_has_merge_conflicts(repo):
            log.debug(f"Merge conflicts:\n{get_merge_conflict_diffs(repo)}")
            if abort_on_conflict:
                print_styled(
                    "Repo has merge conflicts after update, aborting due to abort_on_conflict=True",
                    ALERT_STYLE,
                )
                if cleanup:
                    abort_merge_and_reset_flexlate_branches(
                        repo,
                        current_branch,
                        merged_branch_sha=merged_branch_sha,
                        template_branch_sha=template_branch_sha,
                        merged_branch_name=merged_branch_name,
                        template_branch_name=template_branch_name,
                    )
                raise MergeConflictsAndAbortException

            # Need to wait for user to resolve merge conflicts
            aborted = prompt_to_fix_conflicts_and_reset_on_abort_return_aborted(
                repo,
                current_branch,
                merged_branch_sha,
                template_branch_sha,
                merged_branch_name,
                template_branch_name,
            )
            if aborted:
                raise MergeConflictsAndAbortException

        # No conflicts, merge back into current branch
        current_branch.checkout()
        merge_branch_into_current(repo, merged_branch_name)

        # Current working directory or out root may have been deleted if it was a remove operation
        # and there was nothing else in the folder (git does not save folders without files)
        ensure_exists_folders = [cwd]
        for renderable in orig_renderables:
            ensure_exists_folders.append(
                make_absolute_path_from_possibly_relative_to_another_path(
                    renderable.out_root, project_root
                )
            )
        for folder in ensure_exists_folders:
            if not os.path.exists(folder):
                os.makedirs(folder)

        # Folder may have been deleted again while switching branches, so
        # need to set cwd again
        os.chdir(cwd)

        print_styled("Successfully updated template output", SUCCESS_STYLE)

    def get_updates_for_templates(
        self,
        templates: Sequence[Template],
        data: Optional[Sequence[TemplateData]] = None,
        project_root: Path = Path("."),
        config_manager: ConfigManager = ConfigManager(),
    ) -> List[TemplateUpdate]:
        data = data or []
        all_updates = config_manager.get_no_op_updates(project_root=project_root)
        out_updates: List[TemplateUpdate] = []
        for i, template in enumerate(templates):
            try:
                template_data = data[i]
            except IndexError:
                template_data = {}
            # TODO: more efficient algorithm for matching updates to templates
            for update in all_updates:
                if update.template.name == template.name:
                    update.data = merge_data([template_data], [update.data or {}])[0]
                    update.template.version = template.version
                    update.template.path = template.path
                    out_updates.append(update)
        return out_updates

    def update_passed_templates_to_target_versions(
        self,
        templates: Sequence[Template],
        project_root: Path = Path("."),
        finder: MultiFinder = MultiFinder(),
        config_manager: ConfigManager = ConfigManager(),
    ):
        sources_with_templates = config_manager.get_sources_with_templates(
            templates, project_root=project_root
        )
        for source_with_templates in sources_with_templates:
            # TODO: Separate finding from updating the local version of the repo
            #  If the code failed after the find line but before updating the config,
            #  then the config would not match what is on the disk
            source = source_with_templates.source
            kwargs: Dict[str, Any] = {}
            if source.target_version:
                kwargs.update(version=source.target_version)
            for template in source_with_templates.templates:
                new_template = finder.find(str(source.update_location), **kwargs)
                if template.version != new_template.version:
                    # Template needs to be upgraded
                    # As finder already updates the local files, just update the template object
                    template.version = new_template.version
                    template.path = new_template.path


def _commit_message(renderables: Sequence[Renderable]) -> str:
    message = "Update flexlate templates\n\n"
    for renderable in renderables:
        template = renderable.template
        message += f"{template.name}: {template.version}\n"
    return message


def _create_cwd_and_directories_if_needed(cwd: Path, directories: Sequence[Path]):
    make_all_dirs([cwd])

    # Folder may have been deleted again while switching branches, so
    # need to set cwd again
    os.chdir(cwd)

    make_all_dirs(directories)


def _move_update_config_locations_to_new_parent(
    updates: Sequence[TemplateUpdate], orig_parent: Path, new_parent: Path
) -> List[TemplateUpdate]:
    cwd = Path(os.getcwd())
    return [
        update.copy(
            update=dict(
                config_location=location_relative_to_new_parent(
                    update.config_location, orig_parent, new_parent, cwd
                )
            )
        )
        for update in updates
    ]


def _move_renderable_out_roots_to_new_parent(
    renderables: Sequence[Renderable], orig_parent: Path, new_parent: Path
) -> List[Renderable]:
    return [
        renderable.copy(
            update=dict(
                out_root=location_relative_to_new_parent(
                    renderable.out_root, orig_parent, new_parent, orig_parent
                )
            )
        )
        for renderable in renderables
    ]


def _copy_renderables_skipping_prompts_if_not_in_updates(
    renderables: Sequence[Renderable],
    updates: Sequence[TemplateUpdate],
    project_root: Path,
) -> List[Renderable]:
    update_renderables = [
        update.to_renderable(project_root=project_root) for update in updates
    ]
    return [
        renderable.copy(update=dict(skip_prompts=renderable not in update_renderables))
        for renderable in renderables
    ]



from unittest.mock import patch

from git import Repo

from flexlate import branch_update
from flexlate.bootstrapper import Bootstrapper
from flexlate.config import FlexlateProjectConfig
from tests.fixtures.templated_repo import *
from tests.fixtures.transaction import bootstrap_transaction
from tests.fs_checks import (
    assert_template_source_cookiecutter_one_added_correctly,
    assert_cookiecutter_one_applied_template_added_correctly,
    assert_project_config_is_correct,
)


def test_bootstrap_cookiecutter_one(
    repo_with_cookiecutter_one_applied_but_no_flexlate: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    bootstrap_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_applied_but_no_flexlate
    template = cookiecutter_one_template

    bootstrapper = Bootstrapper()
    bootstrapper.bootstrap_flexlate_init_from_existing_template(
        repo, template, bootstrap_transaction, no_input=True, data=dict(a="b", c="")
    )

    # Check that all Flexlate config files are correct
    assert_template_source_cookiecutter_one_added_correctly(cookiecutter_one_template)
    assert_cookiecutter_one_applied_template_added_correctly(
        template, GENERATED_REPO_DIR / "b"
    )
    assert_project_config_is_correct()

    _assert_flexlate_merge_branch_exists_and_is_up_to_date(repo)


def test_bootstrap_cookiecutter_one_with_conflicts(
    repo_with_cookiecutter_one_applied_but_no_flexlate: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    bootstrap_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_applied_but_no_flexlate
    template = cookiecutter_one_template

    # Modify templated output to cause conflict
    content_path = GENERATED_REPO_DIR / "b" / "text.txt"
    content_path.write_text("merge conflict")
    stage_and_commit_all(
        repo, "Add a change that should cause a merge conflict on bootstrap"
    )

    bootstrapper = Bootstrapper()

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        stage_and_commit_all(repo, "Resolve conflicts")
        return True

    with patch.object(branch_update, "confirm_user", _resolve_conflicts_then_type_yes):
        bootstrapper.bootstrap_flexlate_init_from_existing_template(
            repo, template, bootstrap_transaction, no_input=True, data=dict(a="b", c="")
        )

    # Check that all Flexlate config files are correct
    assert_template_source_cookiecutter_one_added_correctly(cookiecutter_one_template)
    assert_cookiecutter_one_applied_template_added_correctly(
        template, GENERATED_REPO_DIR / "b"
    )
    assert_project_config_is_correct()

    _assert_flexlate_merge_branch_exists_and_is_up_to_date(repo)


def _assert_flexlate_merge_branch_exists_and_is_up_to_date(repo: Repo):
    master = repo.active_branch
    merged_branch = repo.branches[DEFAULT_MERGED_BRANCH_NAME]  # type: ignore
    assert merged_branch.commit.hexsha == master.commit.hexsha



from pathlib import Path
from typing import Optional

from git import Repo

from flexlate.add_mode import AddMode
from flexlate.branch_update import modify_files_via_branches_and_temp_repo
from flexlate.config_manager import ConfigManager
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.transactions.transaction import (
    create_transaction_commit_message,
    FlexlateTransaction,
)


class UserConfigManager:
    """
    A higher-level version of the config manager that also works with flexlate branches

    ConfigManager is the lower-level version that does not care about branches
    """

    def update_template_source_target_version(
        self,
        name: str,
        target_version: Optional[str],
        repo: Repo,
        transaction: FlexlateTransaction,
        project_path: Path = Path("."),
        add_mode: Optional[AddMode] = None,
        merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        base_merged_branch_name: str = DEFAULT_MERGED_BRANCH_NAME,
        template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        base_template_branch_name: str = DEFAULT_TEMPLATE_BRANCH_NAME,
        remote: str = "origin",
        config_manager: ConfigManager = ConfigManager(),
    ):

        if add_mode == AddMode.USER:
            # No need to use git if adding for user
            config_manager.update_template_source_version(
                name, target_version=target_version, project_root=project_path
            )
            return

        commit_message = create_transaction_commit_message(
            _update_target_version_commit_message(name, target_version),
            transaction,
        )

        # Local or project config, add in git
        modify_files_via_branches_and_temp_repo(
            lambda temp_path: config_manager.update_template_source_version(
                name,
                target_version=target_version,
                project_root=temp_path,
            ),
            repo,
            commit_message,
            project_path,
            merged_branch_name=merged_branch_name,
            base_merged_branch_name=base_merged_branch_name,
            template_branch_name=template_branch_name,
            base_template_branch_name=base_template_branch_name,
            remote=remote,
        )


def _update_target_version_commit_message(
    name: str, target_version: Optional[str]
) -> str:
    display_target_version: str = str(target_version)
    if target_version is None:
        display_target_version = "null"
    return (
        f"Changed target version for template source {name} to {display_target_version}"
    )



import json
import os

import pytest

from flexlate.add_mode import AddMode
from flexlate.adder import Adder
from flexlate.branch_update import get_flexlate_branch_name
from flexlate.config import FlexlateConfig
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.ext_git import delete_local_branch
from flexlate.remover import Remover
from flexlate.transactions.transaction import FlexlateTransaction
from flexlate.update.main import Updater
from tests.config import COOKIECUTTER_ONE_NAME
from tests.fileutils import preprend_cookiecutter_one_generated_text

from tests.fixtures.git import *
from tests.fixtures.template import *
from tests.fixtures.transaction import (
    add_source_transaction,
    add_output_transaction,
    remove_source_transaction,
    remove_output_transaction,
    update_transaction,
)
from tests.gitutils import rename_branch


@pytest.fixture
def repo_with_cookiecutter_one_template_source(
    repo_with_placeholder_committed: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_placeholder_committed
    template = cookiecutter_one_template

    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.add_template_source(
            repo,
            template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )
    yield repo


@pytest.fixture
def repo_with_cookiecutter_remote_version_one_template_source(
    repo_with_placeholder_committed: Repo,
    cookiecutter_remote_version_one_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_placeholder_committed
    template = cookiecutter_remote_version_one_template

    adder = Adder()
    adder.add_template_source(
        repo,
        template,
        add_source_transaction,
        out_root=GENERATED_REPO_DIR,
        target_version=COOKIECUTTER_REMOTE_VERSION_1,
    )
    yield repo


@pytest.fixture
def repo_with_cookiecutter_remote_version_one_template_source_and_no_target_version(
    repo_with_cookiecutter_remote_version_one_template_source: Repo,
) -> Repo:
    repo = repo_with_cookiecutter_remote_version_one_template_source

    # Remove target version, allowing the remote version to update
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    ts = config.template_sources[0]
    ts.target_version = None
    config.save()
    stage_and_commit_all(repo, "Remove target version")

    yield repo


@pytest.fixture
def repo_with_cookiecutter_remote_version_one_template_source_that_will_have_merge_conflict_on_flexlate_operation(
    repo_with_cookiecutter_remote_version_one_template_source: Repo,
) -> Repo:
    repo = repo_with_cookiecutter_remote_version_one_template_source

    # Force a merge conflict by reformatting flexlate config
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config_path.write_text(json.dumps(json.loads(config_path.read_text()), indent=4))
    stage_and_commit_all(repo, "Reformat flexlate config")

    yield repo


@pytest.fixture
def repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation(
    repo_with_template_branch_from_cookiecutter_remote: Repo,
) -> Repo:
    repo = repo_with_template_branch_from_cookiecutter_remote

    # Force a merge conflict by reformatting flexlate config
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    config_path.write_text(json.dumps(json.loads(config_path.read_text()), indent=4))
    stage_and_commit_all(repo, "Reformat flexlate config")

    yield repo


@pytest.fixture
def repo_with_copier_output_subdir_template_source(
    repo_with_placeholder_committed: Repo,
    copier_output_subdir_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_placeholder_committed
    template = copier_output_subdir_template

    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.add_template_source(
            repo,
            template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )
    yield repo


@pytest.fixture
def repo_with_gitignore_and_cookiecutter_one_template_source(
    repo_with_gitignore: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_gitignore
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.add_template_source(
            repo,
            cookiecutter_one_template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )
    yield repo


@pytest.fixture
def repo_with_gitignore_and_cookiecutter_one_template_source_in_repo(
    repo_with_gitignore: Repo,
    cookiecutter_one_template_in_repo: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_gitignore
    # While the same changes are in cookiecutter_one_template_in_repo, they do not persist,
    # so make them again
    new_folder = GENERATED_REPO_DIR / "templates"
    new_folder.mkdir(parents=True)
    new_path = new_folder / COOKIECUTTER_ONE_DIR.name
    shutil.copytree(COOKIECUTTER_ONE_DIR, new_path)
    stage_and_commit_all(repo, "Add cookiecutter one template in repo")
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.add_template_source(
            repo,
            cookiecutter_one_template_in_repo,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )
    yield repo


@pytest.fixture
def repo_with_remote_cookiecutter_template_source(
    repo_with_placeholder_committed: Repo,
    cookiecutter_remote_template: CookiecutterTemplate,
    add_source_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.add_template_source(
            repo,
            cookiecutter_remote_template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )
    yield repo


@pytest.fixture
def repo_with_template_branch_from_cookiecutter_one(
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_cookiecutter_one_template_source
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_one_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
        )
    yield repo


@pytest.fixture
def repo_with_template_branch_from_cookiecutter_remote(
    repo_with_cookiecutter_remote_version_one_template_source: Repo,
    cookiecutter_remote_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_cookiecutter_remote_version_one_template_source
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_remote_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            add_mode=AddMode.PROJECT,
            no_input=True,
        )
    yield repo


@pytest.fixture
def repo_with_template_branch_from_cookiecutter_one_and_feature_flexlate_branches(
    repo_with_template_branch_from_cookiecutter_one: Repo,
) -> Repo:
    repo = repo_with_template_branch_from_cookiecutter_one
    # Simulate conditions of developing on feature branch
    feature_merged_branch_name = get_flexlate_branch_name(
        repo, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name(
        repo, DEFAULT_TEMPLATE_BRANCH_NAME
    )
    rename_branch(repo, DEFAULT_MERGED_BRANCH_NAME, feature_merged_branch_name)
    rename_branch(repo, DEFAULT_TEMPLATE_BRANCH_NAME, feature_template_branch_name)
    yield repo


@pytest.fixture
def repo_with_template_branch_from_cookiecutter_one_project_add_mode(
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_cookiecutter_one_template_source
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_one_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
            add_mode=AddMode.PROJECT,
        )
    yield repo


@pytest.fixture
def repo_with_template_branch_from_cookiecutter_remote_version_one(
    repo_with_cookiecutter_remote_version_one_template_source: Repo,
    cookiecutter_remote_version_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_cookiecutter_remote_version_one_template_source
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_remote_version_one_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
            add_mode=AddMode.PROJECT,
        )
    yield repo


@pytest.fixture
def repo_with_gitignore_and_template_branch_from_cookiecutter_one(
    repo_with_gitignore_and_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_gitignore_and_cookiecutter_one_template_source
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_one_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
        )
    yield repo


@pytest.fixture
def repo_with_gitignore_and_template_branch_from_cookiecutter_one_in_repo(
    repo_with_gitignore_and_cookiecutter_one_template_source_in_repo: Repo,
    cookiecutter_one_template_in_repo: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_gitignore_and_cookiecutter_one_template_source_in_repo
    with change_directory_to(GENERATED_REPO_DIR):
        adder = Adder()
        adder.apply_template_and_add(
            repo,
            cookiecutter_one_template_in_repo,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
        )
    yield repo


@pytest.fixture
def repo_from_cookiecutter_one_with_modifications(
    repo_with_template_branch_from_cookiecutter_one: Repo,
) -> Repo:
    repo = repo_with_template_branch_from_cookiecutter_one
    with change_directory_to(GENERATED_REPO_DIR):
        preprend_cookiecutter_one_generated_text("hello\n")
        stage_and_commit_all(repo, "Prepend cookiecutter text with hello")
    yield repo


@pytest.fixture
def repo_after_updating_cookiecutter_one(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_REPO_DIR
    )
    updater.update(repo, template_updates, update_transaction, no_input=True)
    yield repo


@pytest.fixture
def repo_with_cookiecutter_one_template_source_and_output(
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_one_template: CookiecutterTemplate,
    add_output_transaction: FlexlateTransaction,
) -> Repo:
    repo = repo_with_cookiecutter_one_template_source
    adder = Adder()
    with change_directory_to(GENERATED_REPO_DIR):
        adder.apply_template_and_add(
            repo, cookiecutter_one_template, add_output_transaction, no_input=True
        )
    yield repo


@pytest.fixture
def repo_with_template_source_removed(
    repo_with_cookiecutter_one_template_source: Repo,
    remove_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    remover = Remover()
    with change_directory_to(GENERATED_REPO_DIR):
        remover.remove_template_source(
            repo, COOKIECUTTER_ONE_NAME, remove_source_transaction
        )
    yield repo


@pytest.fixture
def repo_with_applied_output_removed(
    repo_with_template_branch_from_cookiecutter_one: Repo,
    remove_output_transaction: FlexlateTransaction,
):
    repo = repo_with_template_branch_from_cookiecutter_one
    remover = Remover()
    with change_directory_to(GENERATED_REPO_DIR):
        remover.remove_applied_template_and_output(
            repo, COOKIECUTTER_ONE_NAME, remove_output_transaction
        )
    yield repo


@pytest.fixture
def repo_with_cookiecutter_one_applied_but_no_flexlate(
    repo_with_template_branch_from_cookiecutter_one: Repo,
) -> Repo:
    repo = repo_with_template_branch_from_cookiecutter_one
    ts_config_path = GENERATED_REPO_DIR / "flexlate.json"
    os.remove(ts_config_path)
    at_config_path = GENERATED_REPO_DIR / "b" / "flexlate.json"
    os.remove(at_config_path)
    delete_local_branch(repo, DEFAULT_MERGED_BRANCH_NAME)
    delete_local_branch(repo, DEFAULT_TEMPLATE_BRANCH_NAME)
    stage_and_commit_all(
        repo, "Remove flexlate files, converting this into only a cookiecutter template"
    )
    yield repo



# Integration tests
import contextlib
import os
import shutil
from pathlib import Path
from typing import Optional, Callable
from unittest.mock import patch

import appdirs
from _pytest.capture import CaptureFixture

from flexlate.add_mode import AddMode
from flexlate.branch_update import (
    get_flexlate_branch_name,
    get_flexlate_branch_name_for_feature_branch,
)
from flexlate.config import FlexlateConfig
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.exc import TriedToCommitButNoChangesException, UnnecessarySyncException
from flexlate.ext_git import merge_branch_into_current
from flexlate.main import Flexlate
from tests.config import (
    GENERATED_FILES_DIR,
    COOKIECUTTER_REMOTE_URL,
    COOKIECUTTER_REMOTE_NAME,
)
from tests.fixtures.git import *
from tests.fixtures.subdir_style import *
from tests.fixtures.template import *
from tests.fixtures.cli import *
from tests.fixtures.add_mode import *
from tests.integration.fixtures.template_source import *
from tests.gitutils import (
    assert_main_commit_message_matches,
    checkout_new_branch,
    checkout_existing_branch,
)
from tests.integration.cli_stub import CLIRunnerException, capture_output
from tests.integration.template_source_checks import (
    assert_project_files_are_correct,
    assert_project_files_do_not_exist,
    assert_template_sources_config_is_correct,
    assert_template_sources_config_is_empty,
    assert_applied_templates_config_is_correct,
    assert_applied_templates_config_is_empty,
    assert_config_is_correct,
    assert_project_config_is_correct,
    _get_default_data,
    assert_root_template_source_output_is_correct,
)
from tests.integration.undoables import UNDOABLE_OPERATIONS
from tests.test_pusher import add_local_remote_and_check_branches_on_exit


def test_init_project_and_add_source_and_template(
    flexlates: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
    template_source: TemplateSourceFixture,
):
    fxt = flexlates.flexlate
    no_input = flexlates.type == FlexlateType.APP
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        fxt.add_template_source(template_source.path)
        fxt.apply_template_and_add(
            template_source.name, data=template_source.input_data, no_input=no_input
        )

    assert_project_files_are_correct(
        expect_data=template_source.input_data,
        template_source_type=template_source.type,
        version=template_source.default_version,
    )
    config_relative_root = (
        template_source.evaluated_render_relative_root_in_output_creator(
            template_source.input_data
        )
    )
    assert_config_is_correct(
        at_config_path=GENERATED_REPO_DIR / config_relative_root / "flexlate.json",
        expect_applied_template_root=template_source.expect_local_applied_template_path,
        expect_data=template_source.input_data,
        template_source_type=template_source.type,
        version=template_source.default_version,
        name=template_source.name,
        url=template_source.url,
        path=template_source.path,
        render_relative_root_in_output=template_source.render_relative_root_in_output,
        render_relative_root_in_template=template_source.render_relative_root_in_template,
    )

    project_config_path = GENERATED_REPO_DIR / "flexlate-project.json"
    assert_project_config_is_correct(project_config_path, user=False)


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_init_project_for_user_and_add_source_and_template(
    flexlates: FlexlateFixture,
    add_mode: AddMode,
    repo_with_placeholder_committed: Repo,
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project(user=True, default_add_mode=add_mode)
        fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        fxt.apply_template_and_add(COOKIECUTTER_REMOTE_NAME, no_input=True)

    assert_project_files_are_correct()

    if add_mode == AddMode.USER:
        at_config_root = GENERATED_FILES_DIR
        ts_config_root = GENERATED_FILES_DIR
        template_root = GENERATED_REPO_DIR.absolute()
    elif add_mode == AddMode.PROJECT:
        at_config_root = GENERATED_REPO_DIR
        ts_config_root = GENERATED_REPO_DIR
        template_root = Path(".")
    elif add_mode == AddMode.LOCAL:
        at_config_root = GENERATED_REPO_DIR / "abc"
        ts_config_root = GENERATED_REPO_DIR
        template_root = Path("..")
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    assert_config_is_correct(
        at_config_path=at_config_root / "flexlate.json",
        ts_config_path=ts_config_root / "flexlate.json",
        expect_applied_template_root=template_root,
        expect_add_mode=add_mode,
    )

    project_config_path = GENERATED_FILES_DIR / "flexlate-project.json"
    assert_project_config_is_correct(project_config_path, user=True, add_mode=add_mode)


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_init_project_and_add_source_and_template_in_subdir(
    flexlates: FlexlateFixture,
    add_mode: AddMode,
    subdir_style: SubdirStyle,
    repo_with_placeholder_committed: Repo,
    template_source_one_remote_and_all_local_relative: TemplateSourceFixture,
):
    fxt = flexlates.flexlate
    template_source = template_source_one_remote_and_all_local_relative
    repo = repo_with_placeholder_committed

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project(default_add_mode=add_mode)
        fxt.add_template_source(template_source.path)
        subdir = GENERATED_REPO_DIR / "subdir1" / "subdir2"
        subdir.mkdir(parents=True)
        if subdir_style == SubdirStyle.CD:
            with change_directory_to(subdir):
                fxt.apply_template_and_add(template_source.name, no_input=True)
        elif subdir_style == SubdirStyle.PROVIDE_RELATIVE:
            fxt.apply_template_and_add(
                template_source.name,
                no_input=True,
                out_root=subdir.relative_to(os.getcwd()),
            )
        elif subdir_style == SubdirStyle.PROVIDE_ABSOLUTE:
            fxt.apply_template_and_add(
                template_source.name, no_input=True, out_root=subdir.absolute()
            )

    assert_project_files_are_correct(
        subdir,
        template_source_type=template_source.type,
        version=template_source.default_version,
    )

    config_relative_root = (
        template_source.evaluated_render_relative_root_in_output_creator(
            _get_default_data(template_source.type)
        )
    )

    if add_mode == AddMode.LOCAL:
        applied_config_dir = subdir / config_relative_root
        expect_applied_template_root = (
            template_source.expect_local_applied_template_path
        )
        template_sources_config_dir = GENERATED_REPO_DIR
        expect_template_source_path = template_source.path
    elif add_mode == AddMode.PROJECT:
        applied_config_dir = GENERATED_REPO_DIR
        expect_applied_template_root = subdir.relative_to(GENERATED_REPO_DIR)
        template_sources_config_dir = GENERATED_REPO_DIR
        expect_template_source_path = template_source.path
    elif add_mode == AddMode.USER:
        applied_config_dir = GENERATED_FILES_DIR
        expect_applied_template_root = subdir.absolute()
        template_sources_config_dir = GENERATED_FILES_DIR
        if template_source.is_local_template:
            # Move the original directory down a level in the relative path
            # E.g. move ../../input_files/templates/cookiecutters/one to ../input_files/templates/cookiecutters/one
            expect_template_source_path = str(
                Path(*Path(template_source.path).parts[1:])
            )
        else:
            expect_template_source_path = template_source.path
    else:
        raise ValueError(f"unsupported add mode {add_mode}")

    assert_template_sources_config_is_correct(
        template_sources_config_dir / "flexlate.json",
        version=template_source.default_version,
        name=template_source.name,
        url=template_source.url,
        path=expect_template_source_path,
        render_relative_root_in_output=template_source.render_relative_root_in_output,
        render_relative_root_in_template=template_source.render_relative_root_in_template,
    )
    assert_applied_templates_config_is_correct(
        applied_config_dir / "flexlate.json",
        expect_applied_template_root=expect_applied_template_root,
        version=template_source.default_version,
        template_source_type=template_source.type,
        name=template_source.name,
        expect_add_mode=add_mode,
    )

    project_config_path = GENERATED_REPO_DIR / "flexlate-project.json"
    assert_project_config_is_correct(project_config_path, user=False, add_mode=add_mode)


@pytest.mark.parametrize("update_from_subdir", [False, True])
def test_update_project(
    update_from_subdir: bool,
    flexlates: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
    template_source_with_relative: TemplateSourceFixture,
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed
    template_source = template_source_with_relative

    no_input = flexlates.type == FlexlateType.APP
    subdir = GENERATED_REPO_DIR / "subdir1" / "subdir2"
    subdir.mkdir(parents=True)

    def assert_root_template_output_is_correct(
        after_version_update: bool = False,
        after_data_update: bool = False,
        target_version: Optional[str] = None,
    ):
        assert_root_template_source_output_is_correct(
            template_source,
            after_version_update,
            after_data_update,
            target_version=target_version,
        )

    def assert_subdir_template_output_is_correct(
        after_version_update: bool = False, after_data_update: bool = False
    ):
        version = (
            template_source.version_2
            if after_version_update
            else template_source.version_1
        )
        input_data = (
            template_source.update_input_data
            if after_data_update
            else template_source.input_data
        )
        at_config_path = (
            subdir
            / template_source.evaluated_render_relative_root_in_output_creator(
                input_data
            )
            / "flexlate.json"
        )
        assert_project_files_are_correct(
            subdir,
            expect_data=input_data,
            version=version,
            template_source_type=template_source.type,
        )
        assert_applied_templates_config_is_correct(
            at_config_path,
            expect_applied_template_root=template_source.expect_local_applied_template_path,
            expect_data=input_data,
            version=version,
            template_source_type=template_source.type,
            name=template_source.name,
        )
        assert_template_sources_config_is_empty(at_config_path)

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        fxt.add_template_source(
            template_source.path, target_version=template_source.version_1
        )
        # Add an output in the main directory
        fxt.apply_template_and_add(
            template_source.name, data=template_source.input_data, no_input=no_input
        )
        assert_root_template_output_is_correct(target_version=template_source.version_1)

        # Add an output of the same template source in a subdirectory
        with change_directory_to(subdir):
            fxt.apply_template_and_add(
                template_source.name, data=template_source.input_data, no_input=no_input
            )
            assert_subdir_template_output_is_correct()

        update_directory = subdir if update_from_subdir else GENERATED_REPO_DIR
        with change_directory_to(update_directory):
            # First update does nothing, because version is at target version
            # When using app, it will throw an error
            if flexlates.type == FlexlateType.APP:
                with pytest.raises(TriedToCommitButNoChangesException) as excinfo:
                    fxt.update(no_input=True)
                assert "update did not make any new changes" in str(excinfo.value)
            else:
                # When using CLI stub, it will throw a CLIRunnerException
                with pytest.raises(CLIRunnerException) as excinfo:
                    fxt.update(no_input=True)
                assert "update did not make any new changes" in str(excinfo.value)

            assert_root_template_output_is_correct(
                target_version=template_source.version_1
            )
            assert_subdir_template_output_is_correct()

            # Now update by just passing new data, should change the output
            # even though the version has not changed
            fxt.update(data=[template_source.update_input_data] * 2, no_input=no_input)
            assert_root_template_output_is_correct(
                after_data_update=True, target_version=template_source.version_1
            )
            assert_subdir_template_output_is_correct(after_data_update=True)

            # Now update the target version
            fxt.update_template_source_target_version(
                template_source.name, template_source.version_2
            )

            # Make changes to update local templates to new version (no-op for remote templates)
            template_source.migrate_version(template_source.url_or_absolute_path)

            # Now update should go to new version
            fxt.update(no_input=True)

    assert_root_template_output_is_correct(
        after_version_update=True,
        after_data_update=True,
        target_version=template_source.version_2,
    )
    assert_subdir_template_output_is_correct(
        after_version_update=True, after_data_update=True
    )

    project_config_path = GENERATED_REPO_DIR / "flexlate-project.json"
    assert_project_config_is_correct(project_config_path, user=False)


def test_update_one_template(
    flexlates: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed
    no_input = flexlates.type == FlexlateType.APP

    non_update_template_source: TemplateSourceFixture = COOKIECUTTER_REMOTE_FIXTURE
    with template_source_with_temp_dir_if_local_template(
        COPIER_LOCAL_FIXTURE
    ) as template_source:
        with change_directory_to(GENERATED_REPO_DIR):
            fxt.init_project()
            # Add both template sources and outputs in the main directory at version 1
            for i, ts in enumerate([template_source, non_update_template_source]):
                fxt.add_template_source(ts.path, target_version=ts.version_1)
                fxt.apply_template_and_add(
                    ts.name, data=ts.input_data, no_input=no_input
                )
                num_template_sources = i + 1
                assert_root_template_source_output_is_correct(
                    ts,
                    num_template_sources=num_template_sources,
                    template_source_index=i,
                    target_version=ts.version_1,
                )

            # First update does nothing, because version is at target version
            # When using app, it will throw an error
            if flexlates.type == FlexlateType.APP:
                with pytest.raises(TriedToCommitButNoChangesException) as excinfo:
                    fxt.update([template_source.name], no_input=True)
                assert "update did not make any new changes" in str(excinfo.value)
            else:
                # When using CLI stub, it will throw a CLIRunnerException
                with pytest.raises(CLIRunnerException) as excinfo:
                    fxt.update([template_source.name], no_input=True)
                assert "update did not make any new changes" in str(excinfo.value)

            assert_root_template_source_output_is_correct(
                template_source,
                target_version=template_source.version_1,
                num_template_sources=2,
            )
            assert_root_template_source_output_is_correct(
                non_update_template_source,
                template_source_index=1,
                target_version=non_update_template_source.version_1,
                num_template_sources=2,
            )

            # Now update by just passing new data, should change the output
            # even though the version has not changed
            fxt.update(
                [template_source.name],
                data=[template_source.update_input_data],
                no_input=no_input,
            )
            assert_root_template_source_output_is_correct(
                template_source,
                after_data_update=True,
                target_version=template_source.version_1,
                num_template_sources=2,
            )
            # Other template unaffected
            assert_root_template_source_output_is_correct(
                non_update_template_source,
                template_source_index=1,
                target_version=non_update_template_source.version_1,
                num_template_sources=2,
            )

            # Now update the target version
            fxt.update_template_source_target_version(
                template_source.name, template_source.version_2
            )

            # Make changes to update local templates to new version (no-op for remote templates)
            template_source.migrate_version(template_source.url_or_absolute_path)

            # Now update should go to new version
            fxt.update([template_source.name], no_input=True)

        assert_root_template_source_output_is_correct(
            template_source,
            after_version_update=True,
            after_data_update=True,
            num_template_sources=2,
            target_version=template_source.version_2,
        )
        # Other template unaffected
        assert_root_template_source_output_is_correct(
            non_update_template_source,
            template_source_index=1,
            num_template_sources=2,
            target_version=non_update_template_source.version_1,
        )

        project_config_path = GENERATED_REPO_DIR / "flexlate-project.json"
        assert_project_config_is_correct(project_config_path, user=False)


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
@pytest.mark.parametrize("user", [False, True])
def test_remove_template_source(
    user: bool,
    flexlates: FlexlateFixture,
    add_mode: AddMode,
    subdir_style: SubdirStyle,
    repo_with_placeholder_committed: Repo,
):
    repo = repo_with_placeholder_committed
    fxt = flexlates.flexlate
    config_root = (
        GENERATED_FILES_DIR if add_mode == AddMode.USER else GENERATED_REPO_DIR
    )
    project_config_root = GENERATED_FILES_DIR if user else GENERATED_REPO_DIR
    config_path = config_root / "flexlate.json"
    project_config_path = project_config_root / "flexlate-project.json"
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project(user=user, default_add_mode=add_mode)
        fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        assert config_path.exists()
        fxt.remove_template_source(COOKIECUTTER_REMOTE_NAME)
        assert not config_path.exists()

    assert_project_config_is_correct(project_config_path, user=user, add_mode=add_mode)

    # Works for main dir, now try subdir
    subdir = GENERATED_REPO_DIR / "subdir1" / "subdir2"
    subdir.mkdir(parents=True)
    subdir_config_root: Path
    if add_mode == AddMode.USER:
        subdir_config_root = GENERATED_FILES_DIR
    elif add_mode == AddMode.LOCAL:
        subdir_config_root = subdir
    elif add_mode == AddMode.PROJECT:
        subdir_config_root = GENERATED_REPO_DIR
    else:
        raise ValueError("unsupported add mode")
    subdir_config_path = subdir_config_root / "flexlate.json"

    if subdir_style == SubdirStyle.CD:
        with change_directory_to(subdir):
            fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
            assert subdir_config_path.exists()
            fxt.remove_template_source(COOKIECUTTER_REMOTE_NAME)
            assert not subdir_config_path.exists()
    elif subdir_style == SubdirStyle.PROVIDE_RELATIVE:
        out_root = subdir.relative_to(os.getcwd())
        fxt.add_template_source(
            COOKIECUTTER_REMOTE_URL,
            template_root=out_root,
        )
        assert subdir_config_path.exists()
        fxt.remove_template_source(COOKIECUTTER_REMOTE_NAME, template_root=out_root)
        assert not subdir_config_path.exists()
    elif subdir_style == SubdirStyle.PROVIDE_ABSOLUTE:
        out_root = subdir.absolute()
        fxt.add_template_source(
            COOKIECUTTER_REMOTE_URL,
            template_root=out_root,
        )
        assert subdir_config_path.exists()
        fxt.remove_template_source(COOKIECUTTER_REMOTE_NAME, template_root=out_root)
        assert not subdir_config_path.exists()

    assert_project_config_is_correct(project_config_path, user=user, add_mode=add_mode)


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
@pytest.mark.parametrize("user", [False, True])
def test_remove_applied_template(
    user: bool,
    flexlates: FlexlateFixture,
    add_mode: AddMode,
    subdir_style: SubdirStyle,
    repo_with_placeholder_committed: Repo,
):
    fxt = flexlates.flexlate
    no_input = flexlates.type == FlexlateType.APP
    config_root = (
        GENERATED_FILES_DIR if add_mode == AddMode.USER else GENERATED_REPO_DIR
    )
    project_config_root = GENERATED_FILES_DIR if user else GENERATED_REPO_DIR
    config_path = config_root / "flexlate.json"
    project_config_path = project_config_root / "flexlate-project.json"
    expect_data: CookiecutterRemoteTemplateData = dict(name="woo", key="it works")
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project(user=user, default_add_mode=add_mode)
        fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        fxt.apply_template_and_add(
            COOKIECUTTER_REMOTE_NAME, data=expect_data, no_input=no_input
        )
        assert_project_files_are_correct(expect_data=expect_data)
        fxt.remove_applied_template_and_output(COOKIECUTTER_REMOTE_NAME)

        assert_project_files_do_not_exist(expect_data=expect_data)
        assert_project_config_is_correct(
            project_config_path, user=user, add_mode=add_mode
        )
        assert_applied_templates_config_is_empty(config_path)

        # Works for main dir, now try subdir
        subdir = GENERATED_REPO_DIR / "subdir1" / "subdir2"
        subdir.mkdir(parents=True)
        if subdir_style == SubdirStyle.CD:
            with change_directory_to(subdir):
                fxt.apply_template_and_add(
                    COOKIECUTTER_REMOTE_NAME, data=expect_data, no_input=no_input
                )
                fxt.remove_applied_template_and_output(COOKIECUTTER_REMOTE_NAME)
        elif subdir_style == SubdirStyle.PROVIDE_RELATIVE:
            out_root = subdir.relative_to(os.getcwd())
            fxt.apply_template_and_add(
                COOKIECUTTER_REMOTE_NAME,
                data=expect_data,
                no_input=no_input,
                out_root=out_root,
            )
            fxt.remove_applied_template_and_output(
                COOKIECUTTER_REMOTE_NAME, out_root=out_root
            )
        elif subdir_style == SubdirStyle.PROVIDE_ABSOLUTE:
            out_root = subdir.absolute()
            fxt.apply_template_and_add(
                COOKIECUTTER_REMOTE_NAME,
                data=expect_data,
                no_input=no_input,
                out_root=out_root,
            )
            fxt.remove_applied_template_and_output(
                COOKIECUTTER_REMOTE_NAME, out_root=out_root
            )

    assert_project_files_do_not_exist(subdir, expect_data=expect_data)
    assert_project_config_is_correct(project_config_path, user=user, add_mode=add_mode)
    assert not (subdir / "flexlate.json").exists()


@patch.object(appdirs, "user_config_dir", lambda name: GENERATED_FILES_DIR)
def test_undo(
    flexlates: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed
    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        fxt.apply_template_and_add(COOKIECUTTER_REMOTE_NAME, no_input=True)
        # Work in a subdirectory so that we can run all operations.
        # Commit a file so that the folder will persist and ensure that
        # the file is never removed
        subdir = GENERATED_REPO_DIR / "subdir"
        subdir.mkdir()
        subdir_placeholder_path = (subdir / "some-file.txt").resolve()
        subdir_placeholder_path.write_text("something")
        manual_commit_message = "Add a placeholder in a subdir"
        stage_and_commit_all(repo, manual_commit_message)
        # Add an operation that will be undone
        with change_directory_to(subdir):
            # One check being careful about the input files, just to make sure
            # something is happening
            fxt.apply_template_and_add(COOKIECUTTER_REMOTE_NAME, no_input=True)
            config_path = subdir / "abc" / "flexlate.json"
            assert_project_files_are_correct(subdir)
            assert_applied_templates_config_is_correct(
                config_path, expect_applied_template_root=Path("..")
            )
            fxt.undo()
            assert_project_files_do_not_exist(subdir)
            assert not config_path.exists()
            assert subdir_placeholder_path.read_text() == "something"
            # Now check everything just to make sure it can be undone
            is_cli = flexlates.type == FlexlateType.CLI
            for operation in UNDOABLE_OPERATIONS:
                operation.operation(fxt, is_cli)
                fxt.undo(num_operations=operation.num_transactions)
                assert_project_files_do_not_exist(subdir)
                assert not config_path.exists()
                assert subdir_placeholder_path.read_text() == "something"

    def assert_merged_commit_history_is_correct():
        assert_main_commit_message_matches(repo.commit().message, manual_commit_message)
        assert len(repo.commit().parents) == 1
        parent = repo.commit().parents[0]
        assert_main_commit_message_matches(parent.message, "Update flexlate templates")

    assert_merged_commit_history_is_correct()
    assert_project_files_are_correct()
    assert_config_is_correct(
        at_config_path=GENERATED_REPO_DIR / "abc" / "flexlate.json",
        expect_applied_template_root=Path(".."),
    )
    assert_project_config_is_correct()

    merged_branch_name = get_flexlate_branch_name(repo, DEFAULT_MERGED_BRANCH_NAME)
    template_branch_name = get_flexlate_branch_name(repo, DEFAULT_TEMPLATE_BRANCH_NAME)

    for branch_name in [merged_branch_name, template_branch_name]:
        branch = repo.branches[branch_name]  # type: ignore
        branch.checkout()

    merged_branch = repo.branches[merged_branch_name]  # type: ignore
    merged_branch.checkout()
    assert_merged_commit_history_is_correct()

    template_branch = repo.branches[template_branch_name]  # type: ignore
    template_branch.checkout()

    assert_main_commit_message_matches(
        repo.commit().message, "Update flexlate templates"
    )
    assert len(repo.commit().parents) == 1
    parent = repo.commit().parents[0]
    assert_main_commit_message_matches(
        parent.message, "Applied template cookiecutter-simple-example to ."
    )


def test_init_project_from_template(
    flexlates: FlexlateFixture,
    template_source_with_relative: TemplateSourceFixture,
):
    fxt = flexlates.flexlate
    template_source = template_source_with_relative

    no_input = flexlates.type == FlexlateType.APP
    with change_directory_to(GENERATED_FILES_DIR):
        subdir = GENERATED_FILES_DIR / "subdir" / "nested"
        fxt.init_project_from(
            template_source.path,
            path=subdir,
            data=template_source.input_data,
            no_input=no_input,
        )

    relative_root = template_source.evaluated_render_relative_root_in_output_creator(
        template_source.input_data
    )
    root = subdir / relative_root
    project_files_check_root = subdir
    project_folder_name = relative_root.name
    if relative_root == Path("."):
        # When relative root is current directory, init-from creates a new folder
        # with name as passed. The default name is "project"
        project_folder_name = "project"
        root = root / project_folder_name
        project_files_check_root = project_files_check_root / project_folder_name

    assert_project_files_are_correct(
        root=project_files_check_root,
        expect_data=template_source.input_data,
        template_source_type=template_source.type,
        version=template_source.default_version,
    )

    assert_config_is_correct(
        at_config_path=root / "flexlate.json",
        ts_config_path=root / "flexlate.json",
        expect_applied_template_root=template_source.expect_local_applied_template_path,
        expect_data=template_source.input_data,
        template_source_type=template_source.type,
        version=template_source.default_version,
        name=template_source.name,
        url=template_source.url,
        path=template_source.relative_path_relative_to(root),
        render_relative_root_in_output=template_source.render_relative_root_in_output,
        render_relative_root_in_template=template_source.render_relative_root_in_template,
    )

    project_config_path = root / "flexlate-project.json"
    assert_project_config_is_correct(
        project_config_path,
        user=False,
        project_folder_name=project_folder_name,
        project_containing_folder=subdir,
    )

    with change_directory_to(root):
        _assert_sync_is_a_no_op(flexlates)


def test_sync_manually_remove_applied_template(
    flexlates: FlexlateFixture, repo_with_placeholder_committed: Repo
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed

    output_folder = GENERATED_REPO_DIR / "abc"

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        fxt.apply_template_and_add(COOKIECUTTER_REMOTE_NAME, no_input=True)

        assert output_folder.exists()
        shutil.rmtree(output_folder)
        stage_and_commit_all(repo, "Manual change to remove applied template")

        fxt.sync()

    merged_branch_name = get_flexlate_branch_name(repo, DEFAULT_MERGED_BRANCH_NAME)
    template_branch_name = get_flexlate_branch_name(repo, DEFAULT_TEMPLATE_BRANCH_NAME)

    for branch_name in [
        "master",
        merged_branch_name,
        template_branch_name,
    ]:
        branch: Head = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        assert not output_folder.exists()


def test_merge(flexlates: FlexlateFixture, repo_with_placeholder_committed: Repo):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        # Make a dummy change
        dummy_file = GENERATED_REPO_DIR / "something.txt"
        dummy_file.write_text("text")
        stage_and_commit_all(repo, "Add a dummy change to the main branch")
        with _checkout_new_branch_that_merges_back(repo, fxt, "add-source"):
            fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        with _checkout_new_branch_that_merges_back(repo, fxt, "add-template"):
            fxt.apply_template_and_add(COOKIECUTTER_REMOTE_NAME, no_input=True)

    for branch_name in ["master", DEFAULT_MERGED_BRANCH_NAME]:
        branch = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        assert_project_files_are_correct()
        assert_main_commit_message_matches(
            "Merge branch 'flexlate-templates-add-template' into flexlate-output-add-template",
            repo.commit().message,
        )

    template_branch = repo.branches[DEFAULT_TEMPLATE_BRANCH_NAME]  # type: ignore
    template_branch.checkout()
    assert_main_commit_message_matches(
        "Update flexlate templates", repo.commit().message
    )


def test_push(flexlates: FlexlateFixture, repo_with_placeholder_committed: Repo):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        # Make a dummy change
        dummy_file = GENERATED_REPO_DIR / "something.txt"
        dummy_file.write_text("text")
        stage_and_commit_all(repo, "Add a dummy change to the main branch")
        with _checkout_new_branch_that_merges_back(repo, fxt, "add-source"):
            fxt.add_template_source(COOKIECUTTER_REMOTE_URL)
        with _checkout_new_branch_that_merges_back(
            repo, fxt, "add-template", delete=False
        ):
            fxt.apply_template_and_add(COOKIECUTTER_REMOTE_NAME, no_input=True)

        feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
            "add-template", DEFAULT_MERGED_BRANCH_NAME
        )
        feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
            "add-template", DEFAULT_TEMPLATE_BRANCH_NAME
        )
        all_branches = [
            DEFAULT_TEMPLATE_BRANCH_NAME,
            DEFAULT_MERGED_BRANCH_NAME,
            feature_merged_branch_name,
            feature_template_branch_name,
        ]

        with add_local_remote_and_check_branches_on_exit(repo, all_branches):
            fxt.push_feature_flexlate_branches("add-template")
            fxt.push_main_flexlate_branches()


def test_check(
    flexlates_ignore_cli_exceptions: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
    capsys: CaptureFixture,
):
    flexlates = flexlates_ignore_cli_exceptions
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed
    template_source: TemplateSourceFixture = COPIER_REMOTE_FIXTURE

    def _run_check_and_asserter_on_output(
        stdout_asserter: Callable[[str], None], stderr_asserter: Callable[[str], None]
    ):
        # Clear the captured output
        capsys.readouterr()

        # Run the check with no names
        output = capture_output(flexlates, capsys, lambda fxt: fxt.check())

        stdout_asserter(output.stdout)
        stderr_asserter(output.stderr)

        # Run the same check passing the name
        output2 = capture_output(
            flexlates, capsys, lambda fxt: fxt.check(names=[template_source.name])
        )
        stdout_asserter(output2.stdout)
        stderr_asserter(output2.stderr)

    def _assert_no_output(output: str):
        assert not output

    def _assert_template_name_old_version_and_new_version_in_output(output: str):
        assert template_source.name in output
        # Version truncates in small or no terminal so just check for beginning
        assert template_source.version_1[:15] in output
        assert template_source.version_2[:15] in output
        assert "Run fxt update" in output

    def _assert_success_output(output: str):
        assert "up to date" in output
        assert not "Run fxt update" in output

    def assert_no_templates_need_to_be_updated():
        _run_check_and_asserter_on_output(_assert_success_output, _assert_no_output)

    def assert_template_needs_to_be_updated():
        _run_check_and_asserter_on_output(
            _assert_template_name_old_version_and_new_version_in_output,
            _assert_no_output,
        )

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        fxt.add_template_source(
            template_source.path, target_version=template_source.version_1
        )

        assert_no_templates_need_to_be_updated()

        # Now update the target version
        fxt.update_template_source_target_version(
            template_source.name, template_source.version_2
        )

        # Make changes to update local templates to new version (no-op for remote templates)
        template_source.migrate_version(template_source.url_or_absolute_path)

        assert_template_needs_to_be_updated()


def test_bootstrap(
    flexlates: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
    template_source_with_relative: TemplateSourceFixture,
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed
    template_source = template_source_with_relative
    no_input = flexlates.type == FlexlateType.APP

    with change_directory_to(GENERATED_REPO_DIR):
        template_source.render_without_flexlate()
        stage_and_commit_all(
            repo, f"Render files with {template_source.template_type.value}"
        )
        fxt.bootstrap_flexlate_init_from_existing_template(
            template_source.path,
            template_version=template_source.version_1,
            data=template_source.input_data,
            no_input=no_input,
        )

        assert_root_template_source_output_is_correct(template_source)

        # Make changes to update local templates to new version (no-op for remote templates)
        template_source.migrate_version(template_source.url_or_absolute_path)

        # Should be anle to directly update after bootstrap
        fxt.update(data=[template_source.update_input_data], no_input=no_input)

        assert_root_template_source_output_is_correct(
            template_source, after_version_update=True, after_data_update=True
        )


def test_update_target_version(
    flexlates: FlexlateFixture,
    repo_with_placeholder_committed: Repo,
    template_source: TemplateSourceFixture,
):
    fxt = flexlates.flexlate
    repo = repo_with_placeholder_committed

    config_path = GENERATED_REPO_DIR / "flexlate.json"

    def assert_target_version_is(version: Optional[str]):
        config = FlexlateConfig.load(config_path)
        assert len(config.template_sources) == 1
        ts = config.template_sources[0]
        assert ts.target_version == version

    with change_directory_to(GENERATED_REPO_DIR):
        fxt.init_project()
        fxt.add_template_source(template_source.path)
        assert_target_version_is(None)
        fxt.update_template_source_target_version(
            template_source.name, target_version=template_source.version_2
        )
        assert_target_version_is(template_source.version_2)

    last_commit_message = f"Changed target version for template source {template_source.name} to {template_source.version_2}"

    assert_main_commit_message_matches(repo.commit().message, last_commit_message)


@contextlib.contextmanager
def _checkout_new_branch_that_merges_back(
    repo: Repo, fxt: Flexlate, branch_name: str, delete: bool = True
):
    checkout_new_branch(repo, branch_name)
    yield
    fxt.merge_flexlate_branches(delete=delete)
    checkout_existing_branch(repo, "master")
    merge_branch_into_current(repo, branch_name)


def _assert_sync_is_a_no_op(flexlate_fixture: FlexlateFixture):
    fxt = flexlate_fixture.flexlate
    if flexlate_fixture.type == FlexlateType.APP:
        with pytest.raises(UnnecessarySyncException):
            fxt.sync()
    else:
        with pytest.raises(CLIRunnerException) as exc_info:
            fxt.sync()
        assert "Everything is up to date" in str(exc_info.value)



import os
from unittest.mock import patch

from flexlate import branch_update
from flexlate.add_mode import AddMode
from flexlate.config import FlexlateConfig
from flexlate.exc import (
    CannotRemoveTemplateSourceException,
    CannotRemoveAppliedTemplateException,
)
from flexlate.ext_git import repo_has_merge_conflicts
from flexlate.remover import Remover
from flexlate.transactions.transaction import FlexlateTransaction
from tests.config import (
    COOKIECUTTER_ONE_NAME,
    COOKIECUTTER_TWO_NAME,
    COOKIECUTTER_REMOTE_NAME,
)
from tests.fixtures.templated_repo import *
from tests.fixtures.transaction import (
    remove_source_transaction,
    remove_output_transaction,
    add_source_transaction,
)
from tests.gitutils import accept_theirs_in_merge_conflict


def test_remove_template_source(
    repo_with_cookiecutter_one_template_source: Repo,
    remove_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    remover = Remover()
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    with change_directory_to(GENERATED_REPO_DIR):
        assert config_path.exists()
        remover.remove_template_source(
            repo, COOKIECUTTER_ONE_NAME, remove_source_transaction
        )
        assert not config_path.exists()


def test_remove_template_source_when_multiple_exist(
    repo_with_cookiecutter_one_template_source: Repo,
    cookiecutter_two_template: CookiecutterTemplate,
    remove_source_transaction: FlexlateTransaction,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    remover = Remover()
    adder = Adder()
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    with change_directory_to(GENERATED_REPO_DIR):
        adder.add_template_source(
            repo, cookiecutter_two_template, add_source_transaction
        )
        remover.remove_template_source(
            repo, COOKIECUTTER_ONE_NAME, remove_source_transaction
        )

    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 0
    assert len(config.template_sources) == 1
    template_source = config.template_sources[0]
    assert template_source.name == COOKIECUTTER_TWO_NAME


def test_remove_template_source_when_outputs_from_another_source_exist(
    repo_with_cookiecutter_one_template_source_and_output: Repo,
    cookiecutter_two_template: CookiecutterTemplate,
    remove_source_transaction: FlexlateTransaction,
    add_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source_and_output
    remover = Remover()
    adder = Adder()
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    with change_directory_to(GENERATED_REPO_DIR):
        adder.add_template_source(
            repo, cookiecutter_two_template, add_source_transaction
        )
        remover.remove_template_source(
            repo, COOKIECUTTER_TWO_NAME, remove_source_transaction
        )

    # Check source successfully removed
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 0
    assert len(config.template_sources) == 1
    template_source = config.template_sources[0]
    assert template_source.name == COOKIECUTTER_ONE_NAME

    # Check no side effects
    unrelated_config_path = GENERATED_REPO_DIR / "b" / "flexlate.json"
    config = FlexlateConfig.load(unrelated_config_path)
    assert len(config.template_sources) == 0
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.add_mode == AddMode.LOCAL
    assert at.version == COOKIECUTTER_ONE_VERSION
    assert at.data == {"a": "b", "c": ""}
    assert at.name == "one"
    assert at.root == Path("..")


def test_remove_non_existing_template_source(
    repo_with_placeholder_committed: Repo,
    remove_source_transaction: FlexlateTransaction,
):
    repo = repo_with_placeholder_committed
    remover = Remover()
    with change_directory_to(GENERATED_REPO_DIR):
        with pytest.raises(CannotRemoveTemplateSourceException) as exc_info:
            remover.remove_template_source(
                repo, COOKIECUTTER_ONE_NAME, remove_source_transaction
            )
        assert "Cannot find any template source" in str(exc_info.value)


def test_remove_template_source_when_output_exists(
    repo_with_cookiecutter_one_template_source_and_output: Repo,
    remove_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source_and_output
    remover = Remover()
    with change_directory_to(GENERATED_REPO_DIR):
        with pytest.raises(CannotRemoveTemplateSourceException) as exc_info:
            remover.remove_template_source(
                repo, COOKIECUTTER_ONE_NAME, remove_source_transaction
            )
        assert "has existing outputs" in str(exc_info.value)


def test_remove_template_source_with_merge_conflict_resolution(
    repo_with_cookiecutter_remote_version_one_template_source_that_will_have_merge_conflict_on_flexlate_operation: Repo,
    remove_source_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_remote_version_one_template_source_that_will_have_merge_conflict_on_flexlate_operation
    remover = Remover()
    config_path = GENERATED_REPO_DIR / "flexlate.json"

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        assert repo_has_merge_conflicts(repo)
        os.remove(config_path)
        stage_and_commit_all(repo, "Manually resolve conflicts")
        return True

    with change_directory_to(GENERATED_REPO_DIR):
        assert config_path.exists()
        with patch.object(
            branch_update, "confirm_user", _resolve_conflicts_then_type_yes
        ):
            remover.remove_template_source(
                repo, COOKIECUTTER_REMOTE_NAME, remove_source_transaction
            )
        assert not config_path.exists()


def test_remove_applied_template(
    repo_with_template_branch_from_cookiecutter_one: Repo,
    remove_output_transaction: FlexlateTransaction,
):
    repo = repo_with_template_branch_from_cookiecutter_one
    remover = Remover()
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    output_path = GENERATED_REPO_DIR / "b" / "text.txt"
    with change_directory_to(GENERATED_REPO_DIR):
        assert output_path.read_text() == "b"
        remover.remove_applied_template_and_output(
            repo, COOKIECUTTER_ONE_NAME, remove_output_transaction
        )

    assert not output_path.exists()
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    assert len(config.applied_templates) == 0


def test_remove_applied_template_that_does_not_exist(
    repo_with_cookiecutter_one_template_source: Repo,
    remove_output_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_one_template_source
    remover = Remover()
    with change_directory_to(GENERATED_REPO_DIR):
        with pytest.raises(CannotRemoveAppliedTemplateException) as exc_info:
            remover.remove_applied_template_and_output(
                repo, COOKIECUTTER_ONE_NAME, remove_output_transaction
            )
        assert "Cannot find any applied template with name" in str(exc_info.value)


def test_remove_applied_template_when_multiple_exist(
    repo_with_cookiecutter_one_template_source_and_output: Repo,
    remove_output_transaction: FlexlateTransaction,
    add_output_transaction: FlexlateTransaction,
    cookiecutter_one_template: CookiecutterTemplate,
):
    repo = repo_with_cookiecutter_one_template_source_and_output
    remover = Remover()
    adder = Adder()
    subdir = GENERATED_REPO_DIR / "subdir"
    subdir.mkdir()
    config_path = subdir / "b" / "flexlate.json"
    output_path = subdir / "b" / "text.txt"
    with change_directory_to(subdir):
        adder.apply_template_and_add(
            repo, cookiecutter_one_template, add_output_transaction, no_input=True
        )
        assert output_path.read_text() == "b"
        assert config_path.exists()
        remover.remove_applied_template_and_output(
            repo, COOKIECUTTER_ONE_NAME, remove_output_transaction
        )

    # Ensure that remove went correctly
    assert not output_path.exists()
    assert not config_path.exists()

    # Ensure no side effects
    unrelated_config_path = GENERATED_REPO_DIR / "b" / "flexlate.json"
    config = FlexlateConfig.load(unrelated_config_path)
    assert len(config.template_sources) == 0
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.add_mode == AddMode.LOCAL
    assert at.version == COOKIECUTTER_ONE_VERSION
    assert at.data == {"a": "b", "c": ""}
    assert at.name == "one"
    assert at.root == Path("..")


def test_remove_applied_template_with_merge_conflict_resolution(
    repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation: Repo,
    remove_output_transaction: FlexlateTransaction,
):
    repo = repo_with_cookiecutter_remote_version_one_template_source_and_output_that_will_have_merge_conflict_on_flexlate_operation
    remover = Remover()
    config_path = GENERATED_REPO_DIR / "flexlate.json"
    output_path = GENERATED_REPO_DIR / "abc" / "abc.txt"

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        assert repo_has_merge_conflicts(repo)
        accept_theirs_in_merge_conflict(repo)
        stage_and_commit_all(repo, "Manually resolve conflicts")
        return True

    with change_directory_to(GENERATED_REPO_DIR):
        assert output_path.read_text() == "some new header\nvalue"
        with patch.object(
            branch_update, "confirm_user", _resolve_conflicts_then_type_yes
        ):
            remover.remove_applied_template_and_output(
                repo,
                COOKIECUTTER_REMOTE_NAME,
                remove_output_transaction,
                add_mode=AddMode.PROJECT,
            )

    assert not output_path.exists()
    config = FlexlateConfig.load(config_path)
    assert len(config.template_sources) == 1
    assert len(config.applied_templates) == 0



import abc
from pathlib import Path
from typing import Union, Protocol, TypeVar

from flexlate.template.base import Template
from flexlate.template_config.base import TemplateConfig

T = TypeVar("T", bound=Template)

# TODO: figure out how to type TemplateFinder properly
class TemplateFinder(Protocol[T]):  # type: ignore
    def find(self, path: str, local_path: Path, **template_kwargs) -> T:
        ...

    def get_config(self, directory: Path) -> TemplateConfig:
        ...

    def matches_template_type(self, path: Path) -> bool:
        ...



import logging
from enum import Enum

from pydantic import BaseSettings, validator
from rich.logging import RichHandler


class LogLevel(str, Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"


class LoggingConfig(BaseSettings):
    level: LogLevel = LogLevel.INFO

    class Config:
        env_prefix = "FLEXLATE_LOG_"

    @validator("level", pre=True)
    def cast_log_level(cls, v):
        if isinstance(v, LogLevel):
            return v
        level = str(v).casefold().strip()
        if level == "info":
            return LogLevel.INFO
        elif level == "debug":
            return LogLevel.DEBUG
        raise ValueError(f"invalid log level {level}")


LOGGING_CONFIG = LoggingConfig()

logging.basicConfig(
    level=LOGGING_CONFIG.level.value,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)

log = logging.getLogger("flexlate")

if __name__ == "__main__":
    log.info("info level")
    log.debug("debug level")
    try:
        raise ValueError("exception")
    except ValueError as e:
        log.exception(e)



import abc
from pathlib import Path
from typing import Optional

from flexlate.template.hashing import md5_dir
from flexlate.template.types import TemplateType
from flexlate.template_config.base import TemplateConfig


class Template(abc.ABC):
    # Override this in subclasses
    _type: TemplateType = TemplateType.BASE

    def __init__(
        self,
        config: TemplateConfig,
        path: Path,
        name: Optional[str] = None,
        version: Optional[str] = None,
        target_version: Optional[str] = None,
        git_url: Optional[str] = None,
        template_source_path: Optional[str] = None,
        render_relative_root_in_output: Path = Path("."),
        render_relative_root_in_template: Path = Path("."),
    ):
        self.config = config
        self.path = path
        self.git_url = git_url
        self.target_version = target_version
        self.name = name or self.default_name
        self.version = version or self.folder_hash
        self.template_source_path = template_source_path or path
        self.render_relative_root_in_output = render_relative_root_in_output
        self.render_relative_root_in_template = render_relative_root_in_template

    @property
    def default_name(self) -> str:
        return self.path.name

    @property
    def folder_hash(self) -> str:
        return md5_dir(self.path)

    def __eq__(self, other):
        try:
            return all(
                [
                    self.config == other.config,
                    self.path == other.path,
                    self.git_url == other.git_url,
                    self.target_version == other.target_version,
                    self.name == other.name,
                    self.version == other.version,
                    self.template_source_path == other.template_source_path,
                    self.render_relative_root_in_output
                    == other.render_relative_root_in_output,
                    self.render_relative_root_in_template
                    == other.render_relative_root_in_template,
                ]
            )
        except AttributeError:
            return False



import re
from pathlib import Path
from typing import Optional, Tuple

import appdirs

from flexlate.exc import InvalidTemplatePathException
from flexlate.ext_git import clone_repo_at_version_get_repo_and_name, checkout_version

CLONED_REPO_FOLDER = Path(appdirs.user_data_dir("flexlate"))

REPO_REGEX = re.compile(
    r"""
# something like git:// ssh:// file:// etc.
((((git|hg)\+)?(git|ssh|file|https?):(//)?)
 |                                      # or
 (\w+@[\w.]+)                          # something like user@...
)
""",
    re.VERBOSE,
)


def is_repo_url(value):
    """Return True if value is a repository URL."""
    return bool(REPO_REGEX.match(value))


def is_local_template(path: str) -> bool:
    return Path(path).exists()


def get_local_repo_path_and_name_cloning_if_repo_url(
    path: str, version: Optional[str] = None, dst_folder: Optional[Path] = None
) -> Tuple[Path, str]:
    if dst_folder is None:
        # Setting here rather than in default params so that mocking
        # during tests is possible
        dst_folder = CLONED_REPO_FOLDER

    if is_local_template(path):
        local_path = Path(path)
        return local_path, local_path.resolve().name

    if not is_repo_url(path):
        raise InvalidTemplatePathException(
            f"Template path {path} is not a valid local path or remote url"
        )

    # Must be a repo url, clone it and return the cloned path
    repo, name = clone_repo_at_version_get_repo_and_name(
        path, dst_folder, version=version
    )

    # For type narrowing
    if repo.working_dir is None:
        raise ValueError("repo working dir cannot be None")
    return Path(repo.working_dir), name



import shutil
import tempfile
from enum import Enum
from pathlib import Path
from typing import Optional, Sequence
from unittest.mock import patch

import pytest
from _pytest.monkeypatch import MonkeyPatch
from git import Repo, Head

from flexlate import branch_update
from flexlate.branch_update import get_flexlate_branch_name_for_feature_branch
from flexlate.config import FlexlateConfig, TemplateSource, TemplateSourceWithTemplates
from flexlate.exc import GitRepoDirtyException, MergeConflictsAndAbortException
from flexlate.finder.multi import MultiFinder
from flexlate.pusher import Pusher
from flexlate.render.specific import cookiecutter
from flexlate.syncer import Syncer
from flexlate.template.base import Template
from flexlate.template.cookiecutter import CookiecutterTemplate
from flexlate.constants import DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME
from flexlate.template.copier import CopierTemplate
from flexlate.template.types import TemplateType
from flexlate.transactions.transaction import FlexlateTransaction
from flexlate.update.main import Updater
from flexlate.update.template import TemplateUpdate
from tests.config import (
    GENERATED_FILES_DIR,
    COOKIECUTTER_REMOTE_URL,
    COOKIECUTTER_REMOTE_VERSION_1,
    COOKIECUTTER_REMOTE_VERSION_2,
    COOKIECUTTER_ONE_VERSION,
    COOKIECUTTER_ONE_NAME,
)
from tests.fileutils import (
    cookiecutter_one_generated_text_content,
    cookiecutter_two_generated_text_content,
)
from tests.fixtures.git import *
from tests.fixtures.template import *
from tests.fixtures.templated_repo import *
from tests.fixtures.updates import *
from tests.fixtures.local_branch_situation import *
from tests.fixtures.transaction import update_transaction, sync_transaction
from flexlate.ext_git import repo_has_merge_conflicts, delete_local_branch

# TODO: check that config is updated after tests
from tests.gitutils import (
    assert_main_commit_message_matches,
    checkout_new_branch,
    add_local_remote,
)


def test_update_template_dirty_repo(
    cookiecutter_one_update_no_data: TemplateUpdate,
    dirty_repo: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = dirty_repo
    updater = Updater()
    with pytest.raises(GitRepoDirtyException):
        updater.update(
            repo, [cookiecutter_one_update_no_data], update_transaction, no_input=True
        )


def _assert_update_of_cookiecutter_one_modified_template_was_successful(
    repo: Repo,
    main_branch_name: str,
    template_branch_name: str,
    merged_branch_name: str,
):
    main_branch: Head = repo.branches[main_branch_name]  # type: ignore
    merged_branch: Head = repo.branches[merged_branch_name]  # type: ignore
    assert repo.active_branch == main_branch
    assert (
        repo.commit().message
        == f"Merge branch '{template_branch_name}' into {merged_branch_name}\n"
    )
    assert_main_commit_message_matches(
        repo.commit().parents[1].message,
        f"Update flexlate templates\n\none: {COOKIECUTTER_ONE_MODIFIED_VERSION}",
    )
    assert (
        cookiecutter_one_generated_text_content(gen_dir=GENERATED_REPO_DIR)
        == "b and extra"
    )
    assert (GENERATED_REPO_DIR / "ignored" / "ignored.txt").exists()
    assert (GENERATED_REPO_DIR / ".gitignore").exists()
    assert (GENERATED_REPO_DIR / "some-dir" / "placeholder.txt").exists()
    merged_branch.checkout()
    assert repo.active_branch == merged_branch
    assert (
        cookiecutter_one_generated_text_content(gen_dir=GENERATED_REPO_DIR)
        == "b and extra"
    )


def test_update_modify_template(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_REPO_DIR
    )
    updater.update(repo, template_updates, update_transaction, no_input=True)
    _assert_update_of_cookiecutter_one_modified_template_was_successful(
        repo, "master", DEFAULT_TEMPLATE_BRANCH_NAME, DEFAULT_MERGED_BRANCH_NAME
    )


@pytest.mark.parametrize("target_specific_template", [False, True])
def test_update_modify_specific_template(
    target_specific_template: bool,
    cookiecutter_one_modified_template: CookiecutterTemplate,
    copier_output_subdir_template: CopierTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one: Repo,
    update_transaction: FlexlateTransaction,
    add_source_transaction: FlexlateTransaction,
    add_output_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one

    # Add a second template source and output that we are not updating
    adder = Adder()
    with change_directory_to(GENERATED_REPO_DIR):
        adder.add_template_source(
            repo,
            copier_output_subdir_template,
            add_source_transaction,
            out_root=GENERATED_REPO_DIR,
        )
        adder.apply_template_and_add(
            repo,
            copier_output_subdir_template,
            add_output_transaction,
            out_root=GENERATED_REPO_DIR,
            no_input=True,
        )

    updater = Updater()
    update_templates: List[Template]
    if target_specific_template:
        update_templates = [cookiecutter_one_modified_template]
    else:
        update_templates = [
            cookiecutter_one_modified_template,
            copier_output_subdir_template,
        ]
    template_updates = updater.get_updates_for_templates(
        update_templates, project_root=GENERATED_REPO_DIR
    )
    updater.update(repo, template_updates, update_transaction, no_input=True)
    _assert_update_of_cookiecutter_one_modified_template_was_successful(
        repo, "master", DEFAULT_TEMPLATE_BRANCH_NAME, DEFAULT_MERGED_BRANCH_NAME
    )

    # Check to make sure other output was not affected
    output_path = GENERATED_REPO_DIR / "aone.txt"
    assert output_path.read_text() == "atwo"


def test_update_modify_template_with_feature_branches_and_main_branches_are_only_on_remote(
    template_branch_situation: LocalBranchSituation,
    output_branch_situation: LocalBranchSituation,
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_REPO_DIR
    )

    feature_branch = "feature"
    checkout_new_branch(repo, feature_branch)
    feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch, DEFAULT_TEMPLATE_BRANCH_NAME
    )

    # Push to remote and delete local branches, so it will have to fetch from remote to branch off
    add_local_remote(repo)
    pusher = Pusher()
    pusher.push_main_flexlate_branches(repo)
    output_branch_situation.apply(repo, DEFAULT_MERGED_BRANCH_NAME)
    template_branch_situation.apply(repo, DEFAULT_TEMPLATE_BRANCH_NAME)

    updater.update(
        repo,
        template_updates,
        update_transaction,
        no_input=True,
        merged_branch_name=feature_merged_branch_name,
        template_branch_name=feature_template_branch_name,
    )

    # Ensure the output is correct
    _assert_update_of_cookiecutter_one_modified_template_was_successful(
        repo, feature_branch, feature_template_branch_name, feature_merged_branch_name
    )

    # Ensure that flexlate base branches were used properly
    template_branch: Head = repo.branches[feature_template_branch_name]  # type: ignore
    base_template_branch: Head = repo.branches[DEFAULT_TEMPLATE_BRANCH_NAME]  # type: ignore
    assert (
        template_branch.commit.parents[0].hexsha == base_template_branch.commit.hexsha
    )

    output_branch: Head = repo.branches[feature_merged_branch_name]  # type: ignore
    base_output_branch: Head = repo.branches[DEFAULT_MERGED_BRANCH_NAME]  # type: ignore
    assert output_branch.commit.parents[0].hexsha == base_output_branch.commit.hexsha


def test_update_modify_template_with_feature_branches_and_feature_branches_are_only_on_remote(
    template_branch_situation: LocalBranchSituation,
    output_branch_situation: LocalBranchSituation,
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one: Repo,
    update_transaction: FlexlateTransaction,
    sync_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_REPO_DIR
    )

    feature_branch = "feature"
    checkout_new_branch(repo, feature_branch)
    feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch, DEFAULT_TEMPLATE_BRANCH_NAME
    )

    # Create flexlate feature branches and add commits to distinguish from base flexlate branches
    syncer = Syncer()
    config_path = GENERATED_REPO_DIR / "b" / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    at = config.applied_templates[0]
    at.data = dict(a="b", c="d")
    config.save()
    stage_and_commit_all(repo, "Update data for applied template")
    syncer.sync_local_changes_to_flexlate_branches(
        repo,
        sync_transaction,
        template_branch_name=feature_template_branch_name,
        merged_branch_name=feature_merged_branch_name,
    )

    # Push to remote and delete local branches, so it will have to fetch from remote to branch off
    add_local_remote(repo)
    pusher = Pusher()
    pusher.push_main_flexlate_branches(repo)
    pusher.push_feature_flexlate_branches(repo)

    expect_base_merged_branch_sha = repo.branches[feature_merged_branch_name].commit.hexsha  # type: ignore
    expect_base_template_branch_sha = repo.branches[feature_template_branch_name].commit.hexsha  # type: ignore

    output_branch_situation.apply(repo, DEFAULT_MERGED_BRANCH_NAME)
    output_branch_situation.apply(repo, feature_merged_branch_name)
    template_branch_situation.apply(repo, DEFAULT_TEMPLATE_BRANCH_NAME)
    template_branch_situation.apply(repo, feature_template_branch_name)

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        stage_and_commit_all(repo, "Resolve conflicts")
        return True

    with patch.object(branch_update, "confirm_user", _resolve_conflicts_then_type_yes):
        updater.update(
            repo,
            template_updates,
            update_transaction,
            no_input=True,
            merged_branch_name=feature_merged_branch_name,
            template_branch_name=feature_template_branch_name,
        )

    # Ensure that flexlate base branches were used properly
    template_branch: Head = repo.branches[feature_template_branch_name]  # type: ignore
    assert template_branch.commit.parents[0].hexsha == expect_base_template_branch_sha
    merged_branch: Head = repo.branches[feature_merged_branch_name]  # type: ignore
    assert merged_branch.commit.parents[0].hexsha == expect_base_merged_branch_sha


def test_update_modify_data(
    cookiecutter_one_template: CookiecutterTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_template],
        data=[{"a": "d", "c": "e"}],
        project_root=GENERATED_REPO_DIR,
    )
    updater.update(repo, template_updates, update_transaction, no_input=True)
    assert (
        cookiecutter_one_generated_text_content(folder="d", gen_dir=GENERATED_REPO_DIR)
        == "de"
    )
    assert (GENERATED_REPO_DIR / "ignored" / "ignored.txt").exists()
    assert (GENERATED_REPO_DIR / ".gitignore").exists()

    # Original b directory should no longer exist, everything moved to d
    assert not (GENERATED_REPO_DIR / "b").exists()

    config_path = GENERATED_REPO_DIR / "d" / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.data == {"a": "d", "c": "e"}


def test_update_modify_data_for_template_in_repo(
    cookiecutter_one_template_in_repo: CookiecutterTemplate,
    repo_with_gitignore_and_template_branch_from_cookiecutter_one_in_repo: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_with_gitignore_and_template_branch_from_cookiecutter_one_in_repo
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_template_in_repo],
        data=[{"a": "d", "c": "e"}],
        project_root=GENERATED_REPO_DIR,
    )
    updater.update(repo, template_updates, update_transaction, no_input=True)
    assert (
        cookiecutter_one_generated_text_content(folder="d", gen_dir=GENERATED_REPO_DIR)
        == "de"
    )
    assert (GENERATED_REPO_DIR / "ignored" / "ignored.txt").exists()
    assert (GENERATED_REPO_DIR / ".gitignore").exists()

    # Original b directory should no longer exist, everything moved to d
    assert not (GENERATED_REPO_DIR / "b").exists()

    config_path = GENERATED_REPO_DIR / "d" / "flexlate.json"
    config = FlexlateConfig.load(config_path)
    assert len(config.applied_templates) == 1
    at = config.applied_templates[0]
    assert at.data == {"a": "d", "c": "e"}


def test_update_modify_template_conflict(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_from_cookiecutter_one_with_modifications: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_from_cookiecutter_one_with_modifications
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_FILES_DIR
    )
    manual_commit_message = "Manually resolve conflicts"

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        assert repo_has_merge_conflicts(repo)
        stage_and_commit_all(repo, manual_commit_message)
        return True

    with patch.object(branch_update, "confirm_user", _resolve_conflicts_then_type_yes):
        updater.update(repo, template_updates, update_transaction, no_input=True)


@patch.object(
    cookiecutter, "prompt_for_config", lambda context, no_input: {"a": "b", "c": ""}
)
def test_update_modify_template_conflict_with_resolution(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_from_cookiecutter_one_with_modifications: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_from_cookiecutter_one_with_modifications
    manual_commit_message = "Manually resolve conflicts"

    def _resolve_conflicts_then_type_yes(prompt: str) -> bool:
        stage_and_commit_all(repo, manual_commit_message)
        return True

    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_FILES_DIR
    )
    with patch.object(branch_update, "confirm_user", _resolve_conflicts_then_type_yes):
        updater.update(repo, template_updates, update_transaction)
    assert not repo_has_merge_conflicts(repo)

    for branch_name in [
        "master",
        DEFAULT_MERGED_BRANCH_NAME,
    ]:
        branch: Head = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        assert repo.commit().message == manual_commit_message + "\n"


@patch.object(
    cookiecutter, "prompt_for_config", lambda context, no_input: {"a": "b", "c": ""}
)
def test_update_modify_template_conflict_with_reject(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_from_cookiecutter_one_with_modifications: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_from_cookiecutter_one_with_modifications

    def _reject_update(prompt: str) -> bool:
        return False

    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_FILES_DIR
    )
    with patch.object(branch_update, "confirm_user", _reject_update):
        with pytest.raises(MergeConflictsAndAbortException):
            updater.update(repo, template_updates, update_transaction)

    assert repo.commit().message == "Prepend cookiecutter text with hello\n"

    for branch_name in [DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME]:
        branch: Head = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        assert_main_commit_message_matches(
            repo.commit().message, "Update flexlate templates"
        )


@patch.object(
    cookiecutter, "prompt_for_config", lambda context, no_input: {"a": "b", "c": ""}
)
def test_update_modify_template_conflict_with_reject_on_feature_branches(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_from_cookiecutter_one_with_modifications: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_from_cookiecutter_one_with_modifications

    def _reject_update(prompt: str) -> bool:
        return False

    feature_branch = "feature"
    checkout_new_branch(repo, feature_branch)

    feature_merged_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch, DEFAULT_MERGED_BRANCH_NAME
    )
    feature_template_branch_name = get_flexlate_branch_name_for_feature_branch(
        feature_branch, DEFAULT_TEMPLATE_BRANCH_NAME
    )

    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_FILES_DIR
    )
    with patch.object(branch_update, "confirm_user", _reject_update):
        with pytest.raises(MergeConflictsAndAbortException):
            updater.update(
                repo,
                template_updates,
                update_transaction,
                merged_branch_name=feature_merged_branch_name,
                template_branch_name=feature_template_branch_name,
            )

    assert repo.commit().message == "Prepend cookiecutter text with hello\n"

    # Main branches are at original state
    for branch_name in [DEFAULT_MERGED_BRANCH_NAME, DEFAULT_TEMPLATE_BRANCH_NAME]:
        branch: Head = repo.branches[branch_name]  # type: ignore
        branch.checkout()
        assert_main_commit_message_matches(
            repo.commit().message, "Update flexlate templates"
        )

    # Feature branches are now deleted
    for branch_name in [feature_template_branch_name, feature_merged_branch_name]:
        with pytest.raises(IndexError):
            branch = repo.branches[branch_name]  # type: ignore


def test_update_modify_template_conflict_with_abort_resets_state(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_from_cookiecutter_one_with_modifications: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_from_cookiecutter_one_with_modifications
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_FILES_DIR
    )

    with pytest.raises(MergeConflictsAndAbortException):
        updater.update(
            repo,
            template_updates,
            update_transaction,
            no_input=True,
            abort_on_conflict=True,
        )

    # Check that repo is in original state
    assert repo.commit().message == "Prepend cookiecutter text with hello\n"
    assert not repo_has_merge_conflicts(repo)


def test_update_modify_template_conflict_with_abort_and_no_cleanup_leaves_git_merge_conflict(
    cookiecutter_one_modified_template: CookiecutterTemplate,
    repo_from_cookiecutter_one_with_modifications: Repo,
    update_transaction: FlexlateTransaction,
):
    repo = repo_from_cookiecutter_one_with_modifications
    updater = Updater()
    template_updates = updater.get_updates_for_templates(
        [cookiecutter_one_modified_template], project_root=GENERATED_FILES_DIR
    )

    with pytest.raises(MergeConflictsAndAbortException):
        updater.update(
            repo,
            template_updates,
            update_transaction,
            no_input=True,
            abort_on_conflict=True,
            cleanup=False,
        )

    # Check that repo is in state from merge conflict
    assert repo_has_merge_conflicts(repo)
    assert repo.commit().message == "Prepend cookiecutter text with hello\n"


@pytest.mark.parametrize(
    "remote_target_version", [None, "main", COOKIECUTTER_REMOTE_VERSION_2]
)
def test_update_passed_templates_to_newest_versions(
    remote_target_version: Optional[str],
    cookiecutter_one_template: CookiecutterTemplate,
    update_transaction: FlexlateTransaction,
):
    wipe_generated_folder()
    local_template = cookiecutter_one_template
    _modify_cookiecutter_one_template_in_generated_folder(local_template)
    updater = Updater()
    finder = MultiFinder()
    remote_template = finder.find(
        COOKIECUTTER_REMOTE_URL, version=COOKIECUTTER_REMOTE_VERSION_1
    )

    class MockConfigManager:
        def get_sources_with_templates(
            self,
            templates: Sequence[Template],
            project_root: Path = Path("."),
            config: Optional[FlexlateConfig] = None,
        ) -> List[TemplateSourceWithTemplates]:
            return [
                TemplateSourceWithTemplates(
                    source=TemplateSource.from_template(
                        remote_template, target_version=remote_target_version
                    ),
                    templates=[remote_template],
                ),
                TemplateSourceWithTemplates(
                    source=TemplateSource.from_template(local_template),
                    templates=[local_template],
                ),
            ]

    assert remote_template.version == COOKIECUTTER_REMOTE_VERSION_1
    assert local_template.version == COOKIECUTTER_ONE_VERSION
    updater.update_passed_templates_to_target_versions(
        [remote_template, local_template], update_transaction, config_manager=MockConfigManager()  # type: ignore
    )
    assert remote_template.version == COOKIECUTTER_REMOTE_VERSION_2
    assert local_template.version == "471c7dbeea2541fe9a7a558d3aafb6e0"


def test_update_passed_templates_to_newest_versions_but_already_at_targets(
    cookiecutter_one_template: CookiecutterTemplate,
    update_transaction: FlexlateTransaction,
):
    wipe_generated_folder()
    local_template = cookiecutter_one_template
    _modify_cookiecutter_one_template_in_generated_folder(local_template)
    updater = Updater()
    finder = MultiFinder()
    remote_template = finder.find(
        COOKIECUTTER_REMOTE_URL, version=COOKIECUTTER_REMOTE_VERSION_1
    )

    class MockConfigManager:
        def get_sources_with_templates(
            self,
            templates: Sequence[Template],
            project_root: Path = Path("."),
            config: Optional[FlexlateConfig] = None,
        ) -> List[TemplateSourceWithTemplates]:
            return [
                TemplateSourceWithTemplates(
                    source=TemplateSource.from_template(
                        remote_template, target_version=COOKIECUTTER_REMOTE_VERSION_1
                    ),
                    templates=[remote_template],
                ),
                TemplateSourceWithTemplates(
                    # It is actually useless to put target version in a local template
                    # TODO: throw an error when user tries to target version for a local template
                    source=TemplateSource.from_template(
                        local_template, target_version=COOKIECUTTER_ONE_VERSION
                    ),
                    templates=[local_template],
                ),
            ]

    assert remote_template.version == COOKIECUTTER_REMOTE_VERSION_1
    assert local_template.version == COOKIECUTTER_ONE_VERSION
    updater.update_passed_templates_to_target_versions(
        [remote_template, local_template], update_transaction, config_manager=MockConfigManager()  # type: ignore
    )
    assert remote_template.version == COOKIECUTTER_REMOTE_VERSION_1
    # Local template is updated anyway, but it shows the old version due to target version
    assert local_template.version == COOKIECUTTER_ONE_VERSION


def _modify_cookiecutter_one_template_in_generated_folder(template: Template):
    template_folder = GENERATED_FILES_DIR / COOKIECUTTER_ONE_NAME
    shutil.copytree(template.path, template_folder, dirs_exist_ok=True)
    template.path = template_folder
    (template_folder / "some file.txt").touch()
